#!/usr/bin/env python

import os
from shutil import copyfile
import argparse as ap

from pynter.slurm.job_settings import JobSettings

class Automation:

    def __init__(self, job_script_filename = None, status_filename='exit_status.txt', path=None):
        """
        Abstract class that contains methods useful to automize calculations

        Parameters
        ----------
        job_script_filename : (str), optional
            Name of script for job submission. The default is set in "config.yml".
        status_filename : (str), optional
            Name for status file written after job exit. The default is "exit_status.txt".
        path : (str), optional
            Path of calculation to automize. The default is None. If None current work dir is used
        """        
        
        self.job_script_filename = job_script_filename if job_script_filename else JobSettings().filename
        self.status_filename = status_filename
        self.path = path if path else os.getcwd()


    def copy_files_to_next_step_dir(self,*args):
        """
        Copy files to directory of next step

        Parameters
        ----------
        *args : (str or tuple)
            File names to copy. If is a string the same name is used for root and destination,
            if is a tuple first string is root, second is destination.
        """        
        next_step_path = self.get_next_step()
        for file in args:
            if isinstance(file,str):
                copyfile(os.path.join(self.path,file),os.path.join(next_step_path,file))
            elif isinstance(file, tuple):
                copyfile(os.path.join(self.path,file[0]),os.path.join(next_step_path,file[1]))
            else:
                raise ValueError('Arguments must be tuples of strings')                
        return
        
    def get_common_path(self):
        """
        Get path that is in common to all steps. This method returns the path removing the last directory. 
        """    
        return os.path.normpath(self.path.replace(os.path.basename(self.path),''))


    def get_next_step(self):
        """
        Get path that  of next step directory. 
        """         
        steps_dirs = self.get_ordered_steps()
        current_index = steps_dirs.index(self.path)
        if current_index+1  < len(steps_dirs):
            next_step_dir = steps_dirs[current_index + 1]
            return next_step_dir
        else:
            print('Dir:"%s" - No next step found' %(self.path))
        

    def get_ordered_steps(self):
        """
        Get list of all the directories present if the first subdirectory and orders them.
        The steps must have the step number as first character

        Returns
        -------
        steps_dirs : (list)
            List of paths of all steps in order.
        """
        common_path = self.get_common_path()
        steps_dirs = [f.path for f in os.scandir(common_path) if f.is_dir()]
        steps_dirs.sort() #might wanna add option to change sorting key. in this way the rule of first number works
        
        return steps_dirs
    
    
    def submit_job(self,job_script_path=None,job_script_filename=None):
        """
        Submits job by launching 'sbatch' command on OS

        Parameters
        ----------
        job_script_path : (str), optional
            Path were job script is located. The default is None. If None the next step path is used.
        job_script_filename : (str), optional
            Filename of job script. The default is None. If None the attribute 'job_script_filename' is used.
        """
        wdir = os.getcwd()
        
        if job_script_path is None:
            job_script_path = self.get_next_step()
        if job_script_filename is None:
            job_script_filename = self.job_script_filename
            
        job_script_path = os.path.normpath(job_script_path)
        os.chdir(job_script_path)        
        os.system('sbatch %s' %job_script_filename)
        os.chdir(wdir)       
        return



class CommandHandler(Automation):
    """
    Subclass of Automation to handle commands from command line
    """

    def parser_common_args(self):
        """
        Add general commands that can be used for every program 

        Returns
        -------
        parser : ArgumentParser object
        """

        parser = ap.ArgumentParser()
        
        parser.add_argument('-j','--job-script',help='Job script filename, default is "job.sh"',required=False,default=self.job_script_filename,type=str,metavar='',dest='job_script_filename')
        parser.add_argument('-s','--status',help='Write exit status to file, default is "exit_status.txt"',required=False,default=self.status_filename,type=str,metavar='',dest='status_filename')
        parser.add_argument('-e','--error-check',action='store_true',help='Perform error checking. Default is False',required=False,default=False,dest='error_check')
        
        return parser
    
    def vasp_args(self):
        """
        Add arguments that can be used for VASP

        Returns
        -------
        args : ArgumentParser object
            ArgumentParser object with parsed arguments. An attribute named with every argument "dest" is created.
        """
        
        parser = self.parser_common_args()
        # part for VASP
        parser.add_argument('-c','--contcar',action='store_true',help='Copy CONTCAR to POSCAR of next step',required=False,default=False,dest='contcar')
        parser.add_argument('-W','--wavecar',action='store_true',help='Copy WAVECAR to next step',required=False,default=False,dest='wavecar')
        parser.add_argument('-C','--chgcar',action='store_true',help='Copy CHGCAR to next step',required=False,default=False,dest='chgcar')
        parser.add_argument('-K','--check-kpoints',action='store_true',help='Copy WAVECAR and POSCAR only if KPOINTS of next step are the same',required=False,default=False,dest='check_kpoints')        
           
        args = parser.parse_args()
        
        args.status_filename = args.status_filename if args.status_filename != 'None' else None # if status filename argument is 'None' string set attribute to None
        
        # update class attributes with passed arguments
        self.job_script_filename = args.job_script_filename
        self.status_filename = args.status_filename
        
        return args

        

                    
                  
                    
                    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  2 15:05:08 2023

@author: villa
"""
import os

from pymatgen.io.vasp.inputs import Incar,Kpoints,Poscar,Potcar

from pynter.data.datasets import Dataset
from pynter.slurm.job_settings import JobSettings
from pynter.vasp.schemes import AdvancedSchemes


def setup_inputs(subparsers):
    
    parser_schemes = subparsers.add_parser('inputs',help='Create inputs for several calculation schemes')
    subparsers_schemes = parser_schemes.add_subparsers()

    parser_vasp = subparsers_schemes.add_parser('vasp',help='Create inputs for VASP DFT calculations')
    setup_inputs_vasp(parser_vasp)
    
    return


def setup_inputs_vasp(parser):
    parser = parse_common_args(parser)
    parser = parse_vasp_args(parser)
    
    parser.add_argument('-cs','--charge-states',action='append',help='Charged states. Provide a list of integer charge states',required=False,
                        type=int,default=None,metavar='',dest='charge_states')
    
    parser.add_argument('-dp','--dielectric-properties',action='store_true',help='Electronic and ionic parts of the dielectric tensor',required=False,
                        default=False,dest='dielectric_properties')

    parser.add_argument('-ec','--encut-convergence',action='store_true',help='Energy cutoff convergence',required=False,
                        default=False,dest='encut_convergence')
    
    parser.add_argument('-fc','--fractional-charge',action='store_true',help='Fractional charge linearity',required=False,
                        default=False,dest='fractional_charge')
    
    parser.add_argument('-kc','--kpoints-convergence',action='store_true',help='Kpoints convergence',required=False,
                        default=False,dest='kpoints_convergence')
        
    parser.add_argument('-hse','--hse-relaxation',action='store_true',help='5 steps atomic and volume relaxation with HSE',required=False,
                        default=False,dest='hse_relaxation')
    
    parser.add_argument('-pbe','--pbe-relaxation',action='store_true',help='3 steps atomic and volume relaxation with PBE',required=False,
                        default=False,dest='pbe_relaxation')
    
    parser.add_argument('-pbe-es','--pbe-electronic-structure',action='store_true',help='3 steps electronic structure calculations with PBE',required=False,
                        default=False,dest='pbe_electronic_structure')
    
    parser.add_argument('-is','--inputsets',help='Name of a set of inputs from the InputSets class',required=False,type=str,
                        default=None,metavar='',dest='input_sets')
    
    parser.set_defaults(func=create_inputs_vasp)
    return


def create_inputs_vasp(args):
    schemes = get_schemes(args)
    jobs = None
    if args.encut_convergence:
        jobs = schemes.convergence_encut()
    
    elif args.kpoints_convergence:
        jobs = schemes.convergence_kpoints()
        
    elif args.charge_states:
        jobs = schemes.charge_states(args.charge_states,locpot=True)
        
    elif args.dielectric_properties:
        jobs = schemes.dielectric_properties_complete()
        
    elif args.fractional_charge:
        jobs = schemes.fractional_charge_linearity()
        
    elif args.pbe_relaxation:
        jobs = schemes.pbe_vol_relaxation()
        
    elif args.pbe_electronic_structure:
        jobs = schemes.pbe_electronic_structure()
    
    elif args.hse_relaxation:
        jobs = schemes.hse_vol_relaxation()
    
    if jobs:
        ds = Dataset(jobs)
        ds.write_jobs_input()        
 
    elif args.input_sets:
        method = getattr(schemes, args.input_sets)
        job = method()
        job.write_input()
    else:
        raise ValueError('Either method of InputSets or scheme must be provided')
    return
        

def get_schemes(args):    
    if args.poscar:
        poscar = Poscar.from_file(os.path.abspath(args.poscar))
        structure = poscar.structure
    elif args.mp_id:
        from pynter.tools.materials_project import MPDatabase
        structure = MPDatabase(args.mp_id).get_structure()
    else:
        raise ValueError('Either POSCAR or Materials Project ID needs to be provided')

    incar = Incar.from_file(os.path.abspath(args.incar)) if args.incar else None
    kpoints = Kpoints.from_file(os.path.abspath(args.kpoints)) if args.kpoints else None
    potcar = Potcar.from_file(os.path.abspath(args.potcar)) if args.potcar else None
    if args.job_script:
        script_path = os.path.abspath(os.path.dirname(args.job_script))
        job_settings = JobSettings.from_bash_file(path=script_path,filename=os.path.basename(args.job_script))
    else:
        job_settings = None
    
    schemes = AdvancedSchemes(path=args.path,structure=structure,incar_settings=incar,
                               kpoints=kpoints,potcar=potcar,job_settings=job_settings,
                               name=args.name,add_parent_folder=True)
    return schemes


def parse_common_args(parser):
    parser.add_argument('-id','--materials-project-id',help='Materials project ID containing the structure, in case the POSCAR is not provided (default: %(default)s)',
                        required=False,type=str,default=None,metavar='',dest='mp_id')

    parser.add_argument('-n','--name',help='Name of the folder containing input files (default: %(default)s)',required=False,type=str,
                        default='inputs',metavar='',dest='name')
    
    parser.add_argument('-p','--path',help='Path for input files (default: %(default)s)',required=False,type=str,
                        default=os.getcwd(),metavar='',dest='path')
    
    return parser


def parse_vasp_args(parser):

    parser.add_argument('-in','--incar',help='INCAR file with general settings, if not provided default settings are used',required=False,type=str,default=None,metavar='',
                        dest='incar')
    
    parser.add_argument('-kp','--kpoints',help='KPOINTS file, if not provided default settings are used',required=False,type=str,default=None,metavar='',
                        dest='kpoints')
    
    parser.add_argument('-pos','--poscar',help='POSCAR file (default: %(default)s)',required=False,type=str,default=None,metavar='',
                        dest='poscar')
    
    parser.add_argument('-pot','--potcar',help='POTCAR file (default: %(default)s)',required=False,type=str,default=None,metavar='',
                        dest='potcar')

    parser.add_argument('-j','--job-script',help='Job script with input job settings',required=False,type=str,
                        default=None,metavar='',dest='job_script')
    
    return parser  
    





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 14 15:28:02 2023

@author: villa
"""

from pynter.testing.core import PynterTest


class ChempotsTest(PynterTest):
    """
    Provides methods to test Chempots objects
    """
    def assert_Chempots_equal(self,chempots1, chempots2, **kwargs):
        """
        Both Chempots or dictionary objects are allowed as input
        """
        dict1 = chempots1 if type(chempots1) == dict else chempots1.mu
        dict2 = chempots2 if type(chempots2) == dict else chempots2.mu
        self.assert_object_almost_equal(dict1, dict2, **kwargs)
        
        
class ReservoirsTest(ChempotsTest):
    """
    Provides methods to test Reservoirs objects
    """
    def assert_Reservoirs_equal(self,reservoirs1, reservoirs2, check_reference=True, **kwargs):
        """
        Reservoirs dictionary, reference chempots and are_chempots_delta are checked.
        Set check_reference to False to check only reservoirs dictionary.
        """
        keys1 = list(reservoirs1.keys())
        keys2 = list(reservoirs2.keys())
        self.assert_object_almost_equal(keys1, keys2, **kwargs)
        for key in reservoirs1:
            chempots1 = reservoirs1[key]
            chempots2 = reservoirs2[key]
            self.assert_Chempots_equal(chempots1, chempots2, **kwargs)
        if check_reference:
            self.assert_Chempots_equal(reservoirs1.mu_refs, reservoirs2.mu_refs, **kwargs)
            assert reservoirs1.are_chempots_delta == reservoirs2.are_chempots_delta


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 12 15:37:44 2023

@author: lorenzo
"""
from pynter.testing.core import PynterTest

class JobSettingsTest(PynterTest):
    """
    Provides methods to test job settings dictionary
    """
    @property
    def common_keys(self):
        common_keys = [
            'add_automation',
            'add_stop_array',
            'array_size',
            'filename',
            ]
        return common_keys
    
    @property
    def slurm_keys(self):
        slurm_keys =  [
            'ntasks',
            'error',
            'mem-per-cpu',
            'job-name',
            'output',
            'time'
            ]
        return slurm_keys
    
    def assert_job_settings_equal(self,settings1,settings2):
        """
        Check only common keys to avoid system dependent differences
        """
        for key in self.common_keys:
            actual = settings1[key]
            desired = settings2[key]
            self.assert_object_almost_equal(actual,desired)
        
        for kwarg in self.slurm_keys: 
            actual = settings1['slurm'][kwarg]
            desired = settings2['slurm'][kwarg]
            self.assert_object_almost_equal(actual,desired)
        



import os.path as op
from shutil import which

from monty.dev import requires

from pynter import run_local, SETTINGS
    

class HPCInterface:
    
    
    def __init__(self,config=None):
        """
        Class to interface commands with HPC

        Parameters
        ----------
        config : (dict), optional
            Configuration dictionary. If None is loaded from .pynter/config.yml.
        """
        if not config:
            config = SETTINGS['HPC']
        
        for key,value in config.items():
            setattr(self, key, value)      


    def cancel_jobs(self,*args,printout=True,dry_run=False,**kwargs):
        """
        Cancel jobs on HPC using scancel

        Parameters
        ----------
        *args : (str)
            Job-id.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        kwargs : (dict), optional
            Kwargs for the run_local function.
        """
        cmd = 'scancel '
        for arg in args:
            cmd += arg + ' '
                
        stdout,stderr = self.command(cmd,printout,dry_run,**kwargs)
        return stdout,stderr


    @requires(which("sshpass"),
              "sshpass needs to be installed, you can install it with 'sudo apt-get install sshpass'.")   
    
    def command(self,cmd,printout=True,dry_run=False,**kwargs):   
        """
        Run command on HPC

        Parameters
        ----------
        cmd : (str)
            Command to run.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        
        cmd_split = cmd.split()
        arg = ''
        for c in cmd_split:
            arg += '"%s" ' %c
         
        command = 'sshpass ssh %s %s' %(self.hostname, arg)
        if printout:
            print(f'{self.hostname}: {cmd} \n')
        stdout,stderr = run_local(command,printout,dry_run,**kwargs)
        return stdout,stderr
                

    def mkdir(self,path,printout=True,dry_run=False,**kwargs):
        """
        Make new directory in HPC if doesn't exist

        Parameters
        ----------
        path : (str)
            Path of directory.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        cmd = 'mkdir -p %s' %path
        stdout,stderr = self.command(cmd,printout,dry_run,**kwargs)
        return stdout, stderr
        
        
    def qstat(self,cmd='squeue -o "%.10i %.9P %.40j %.8u %.2t %.10M %.5D %R"',printout=True,dry_run=False,**kwargs):
        """
        Check queue status on HPC

        Parameters
        ----------
        cmd : (str), optional
            Command to run. The default is 'squeue -o "%.10i %.9P %.40j %.8u %.2t %.10M %.5D %R"'.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        
        stdout,stderr = self.command(cmd,printout,dry_run,**kwargs)
        return stdout, stderr
    
    
    @requires(which("rsync"),
          "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")
    
    def rsync_from_hpc(self,remotedir=None,localdir=None,exclude=None,printout=True,dry_run=False,**kwargs):
        """
        Sync folders from HPC to local machine. The command "rsync" is used. With this function all
        the folders contained in the remote dir are synched to the local dir.

        Parameters
        ----------
        remotedir : (str), optional
            Remote directory. The default is None. If None the one in config.yml file is used.
        localdir : (str), optional
            Local directory. The default is None. If None the one in config.yml file is used.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """        
        remotedir = remotedir if remotedir else self.workdir
        localdir = localdir if localdir else self.localdir
        localdir = op.abspath(localdir)
        remotedir = op.join(remotedir,'')  #ensure backslash at the end
        localdir = op.join(localdir,'')
        
        localcmd = 'mkdir -p %s' %localdir
        run_local(localcmd)
        
        cmd = "rsync -r -uavzh " #keep the spaces
        if dry_run:
            cmd += "--dry-run "
        if exclude:
            for s in exclude:
                cmd += f'--exclude={s} ' 
        cmd += f"-e ssh {self.hostname}:{remotedir} {localdir} "

        stdout,stderr = run_local(cmd,printout=printout,dry_run=False,**kwargs)
        return stdout,stderr


    @requires(which("rsync"),
          "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")
    
    def rsync_to_hpc(self,localdir=None,remotedir=None,exclude=None,printout=True,dry_run=False,**kwargs):
        """
        Sync folders from local machine to HPC. The command "rsync" is used. With this function all
        the folders contained in the local dir are synched to the remote dir.

        Parameters
        ----------
        localdir : (str), optional
            Local directory. The default is None. If None the one in config.yml file is used.
        remotedir : (str), optional
            Remote directory. The default is None. If None the one in config.yml file is used.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """        
        localdir = localdir if localdir else self.localdir 
        remotedir = remotedir if remotedir else self.workdir
        localdir = op.abspath(localdir)
        localdir = op.join(localdir,'')  #ensure backslash at the end
        remotedir = op.join(remotedir,'')
        
        self.mkdir(remotedir,printout=False)
        
        cmd = "rsync -r -uavzh " #keep the spaces
        if dry_run:
            cmd += "--dry-run "
        if exclude:
            for s in exclude:
                cmd += f'--exclude={s} '
        cmd += f"-e ssh  {localdir} {self.hostname}:{remotedir} "
        
        stdout,stderr = run_local(cmd,printout=printout,dry_run=False,**kwargs)

        return stdout,stderr
    
    
    @requires(which("sshpass"),
              "sshpass needs to be installed, you can install it with 'sudo apt-get install sshpass'.")      
    
    def sbatch(self,path='',job_script_filename='job.sh',printout=True,dry_run=False,**kwargs):
        """
        Execute "sbatch" command on HPC to run job.

        Parameters
        ----------
        path : (str), optional
            Path where the job script is stored. The default is ''.
        job_script_filename : (str), optional
            Filename of the job script. The default is 'job.sh'.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        kwargs : (dict), optional
            Kwargs for the run_local function.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """    
        path = op.join(self.workdir,path)
        cmd = f'cd {path} ; sbatch {job_script_filename}'
        command = f'sshpass ssh {self.hostname} ' + cmd
        stdout,stderr = run_local(command,printout,dry_run,**kwargs) # I've used the run local because there was a probelm with the multiple command given with ;
        return stdout,stderr               # to check again if possible
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 10:52:08 2020

@author: villa
"""

import numpy as np
import os.path as op
import os
import collections
from pymatgen.io.ase import AseAtomsAdaptor
from ase.visualize import view
from pymatgen.util.coord import pbc_shortest_vectors
from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition
from pymatgen.core.trajectory import Trajectory

def _get_distance_vector_and_image(lattice,frac_coords1,frac_coords2,jimage=None):
    """
    Same as pymatgen.core.Lattice.get_distance_and_image but returns 
    the distance vetor instead of the norm.
    Gets distance between two frac_coords assuming periodic boundary
    conditions. If the index jimage is not specified it selects the j
    image nearest to the i atom and returns the distance and jimage
    indices in terms of lattice vector translations. If the index jimage
    is specified it returns the distance between the frac_coords1 and
    the specified jimage of frac_coords2, and the given jimage is also
    returned.
    Args:
        frac_coords1 (3x1 array): Reference fcoords to get distance from.
        frac_coords2 (3x1 array): fcoords to get distance from.
        jimage (3x1 array): Specific periodic image in terms of
            lattice translations, e.g., [1,0,0] implies to take periodic
            image that is one a-lattice vector away. If jimage is None,
            the image that is nearest to the site is found.
    Returns:
        (distance, jimage): distance and periodic lattice translations
        of the other site for which the distance applies. This means that
        the distance between frac_coords1 and (jimage + frac_coords2) is
        equal to distance.
    """
    if jimage is None:
        v, d2 = pbc_shortest_vectors(lattice, frac_coords1, frac_coords2, return_d2=True)
        fc = lattice.get_fractional_coords(v[0][0]) + frac_coords1 - frac_coords2  # type: ignore
        fc = np.array(np.round(fc), dtype=int)
        return v, fc

    jimage = np.array(jimage)
    mapped_vec = lattice.get_cartesian_coords(jimage + frac_coords2 - frac_coords1)  # type: ignore
    return mapped_vec, jimage  # type: ignore


def get_distance_vector(site1,site2,jimage=None):
    """
    Get distance vector between two sites assuming periodic boundary conditions.
    Same as pymatgen.core.sites.Site.distance but returns a vector instead of norm.
    Args:
        other (PeriodicSite): Other site to get distance from.
        jimage (3x1 array): Specific periodic image in terms of lattice
            translations, e.g., [1,0,0] implies to take periodic image
            that is one a-lattice vector away. If jimage is None,
            the image that is nearest to the site is found.
    Returns:
        distance (float): Distance between the two sites
    """
    return _get_distance_vector_and_image(site1.lattice, site1.frac_coords, site2.frac_coords,jimage=jimage)[0]


def get_displacement_vectors(structure1,structure2):
    """
    Get vectors in cartesian coords of site displacements of structure2 w.r.t. structure 1

    Parameters
    ----------
    structure1 : 
        Pymatgen Structure object.
    structure2 : 
        Pymatgen Structure object.

    Returns
    -------
    (numpy ndarray)
        Displacement vectors in cartesian coordinates.
    """
    traj = Trajectory.from_structures([structure1,structure2],constant_lattice=True) #order of structures determines ref in traj
    traj.to_displacements()
    disp = traj.frac_coords[1]
    return structure1.lattice.get_cartesian_coords(disp)


def is_site_in_structure(site,structure,tol=1e-03):
    """
    Check if Site is part of the Structure list. This function is needed because 
    sometimes doing a simple check ("site in structure") doesn't work. This function performes
    a check on the coordinates and the element on the site. Therefore it is more reliable.

    Parameters
    ----------
    site : (Site)
        PeriodicSite or Site object.
    structure : (Structure)
        Pymatgen Structure object.
    tol : (float), optional
        Tolerance for fractional coordinates. The default is 1e-03.

    Returns
    -------
    is_site_in_structure : (bool)
    index : (int)
        Index of site in structure in case site is_site_in_structure returns True
        If False index will be None.
    """
    is_site_in_structure,index = False,None
    check,index = is_site_in_structure_coords(site, structure,tol=tol)
    if check:
        s = structure[index]
        if site.specie.symbol == s.specie.symbol:
            is_site_in_structure =True
    
    return is_site_in_structure,index


def is_site_in_structure_coords(site,structure,tol=1e-03):
    """
    Check if Site coordinates are prensent in the Structure. Calculates distance between target site and 
    each site in reference structure, takes the minimum value and returns True if is within a tolerance.
    The periodicity is accounted for (considering the lattice associated to the PeriodicSite).
    The tolerance is normalized with respect to lattice vector size. 

    Parameters
    ----------
    site : (Site)
        PeriodicSite or Site object.
    structure : (Structure)
        Pymatgen Structure object.
    tol : (float), optional
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.

    Returns
    -------
    is_site_in_structure_coords : (bool)
    index : (int)
        Index of site in structure in case site is_site_in_structure_coords returns True
        If False index will be None.
    """
    l = site.lattice
    tol = np.sqrt(l.a**2 + l.b**2 + l.c**2) * tol #input is normalized with respect to lattice vector
    distances=[]
    for s in structure:
        distances.append(s.distance(site))
    
    distance_min = min(distances)
    if distance_min < tol:
        i_min = distances.index(distance_min)
        index = structure.index(structure[i_min])
        return True, index
    else:
        return False,None


def remove_oxidation_state_from_site(site):
    """
    Remove oxidation state decoration from site.
    """
    new_sp = collections.defaultdict(float)
    for el, occu in site.species.items():
        sym = el.symbol
        new_sp[Element(sym)] += occu
    site.species = Composition(new_sp)
    return

def sort_sites_to_ref_coords(structure,structure_ref,extra_sites=[],tol=1e-03,get_indexes=False):
    """
    Sort Sites of one structure to match the order of coordinates in a reference structure. 

    Parameters
    ----------
    structure : (Structure)
        Structure to sort.
    structure_ref : (Structure)
        Reference Structure.
    extra_sites : (list), optional
        Sites to append at the end of the structure. The default is [].
    tol: (float)
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    get_indexes : (bool), optional
        Get list of mapping indexes for target structure sites in reference structure. The default is False.

    Returns
    -------
    new_structure : (Structure)
        Sorted Structure
    indexes : (list)
        If get_indexes is True a tuple is returned. List of mapping indexes for target structure sites
        in reference structure.
    """
    df = structure
    bk = structure_ref
    indexes = []
    new_sites=[]
    for s in df:
        check,index = is_site_in_structure_coords(s,bk,tol=tol)
        if check:
            indexes.append(index)
    for w in range(0,len(bk)):
        new_sites.insert(indexes[w],df[w])
    for s in extra_sites:
        new_sites.append(s)
            
    new_structure = df.copy()
    new_structure._sites = new_sites 
    if get_indexes:
        return new_structure, indexes
    else:
        return new_structure


def view_structure_with_ase(structures):
    """
    Visualize a Structure object or a list of Structure objects with the ASE. 
    First the Structure objects are converted into an ase Atom object, then "view" is used to visualize them.
    """
    if type(structures) == list:
        atoms=[]
        for s in structures:
            atoms.append(AseAtomsAdaptor.get_atoms(s))
    else:
        atoms = AseAtomsAdaptor.get_atoms(structures)
    view(atoms)
    return


def write_extxyz_file(file,structure,structure_ref=None,displacements=False):
    """
    Write extxyz format. Displacements can be included for visualization in OVITO. 

    Parameters
    ----------
    file : (str)
        Path to save extxyz file to.
    structure : (Structure)
        Structure to visualize.
    structure_ref : (Structure)
        Reference Structure if needed. The default is None.
    displacements : (bool)
        Include displacement vectors. The default is False.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)
    if displacements:
        disp = get_displacement_vectors(structure_ref, structure)
        atoms.arrays.update({'disp':disp})
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    atoms.write(file,format='extxyz')
    return


def write_xdatcar_from_structures(structures,file='XDATCAR'):
    """
    Write XDATCAR file from a list of structures. The first structure determines the reference for the Trajectory object.
    """
    traj = Trajectory.from_structures(structures,constant_lattice=True) 
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    traj.write_Xdatcar(file)
    return
    
    
    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 15:40:26 2020

@author: villa
"""
import pynter
import re
import os
import os.path as op
import json
import pkgutil
import pandas as pd
from monty.json import jsanitize,MontyEncoder, MontyDecoder


def change_file (input_file , output_file=None, back_up_file = True,
                 str_to_modify = {}, lines_to_add=[], check_added_lines = True, lines_to_remove = []):
    
    ''' function to generate a copy of a file changing a string in specific lines
         In case input and output files are the same a copy of the old file is 
         created with name 'old_{input_file}'
    
     input_file: input file to change
     output_file: output file to generate
                  - Default is 'None' - in this case the same is set equal as the input file
                  - In case output_file and input_file have the same name a copy of 
                    the original input file is saved
     str_to_modify: dictionary of strings to be modified - format: {'old string':'new string'}
     lines_to_add: list of lines to be added at the end of the file: format: ['string 1','string 2']
     check_added_lines: - True - if line is already present is not added
                        - False - line is added indipendently of the others
    ''' 
    origin_file = open(input_file,'r')
    lines_list = origin_file.readlines()
    origin_file.close()
    
    if output_file == None or input_file == output_file:
        output_file = input_file
        if back_up_file == True:
            os.rename(input_file, 'old_' + input_file) 
        else:
            os.remove(input_file)
    
    new_file = open(output_file,'w')
           
    for line in lines_list:        
        rem_line = False
        mod_line = False 
        for l in lines_to_remove:
            if l + '\n' == line:
                rem_line = True
        for string in str_to_modify: 
            target_line = re.findall(string, line)                      
            if target_line != [] and rem_line == False:
                new_file.write(line.replace(string,str_to_modify[string]))
                mod_line = True                  
        if mod_line == False and rem_line == False:
             new_file.write(line)   
         
    for l in lines_to_add:
          if check_added_lines:              
              if l + '\n' not in lines_list:            
                  new_file.write(l + '\n')
          else:
              new_file.write(l + '\n')
    
    new_file.close()
    
    return
    

def display_df(df,max_rows=None,max_columns=None,max_colwidth=None):
    """
    Print DataFrame with temporary options

    Parameters
    ----------
    df : (DataFrame)
        DataFrame object.
    max_rows : (int), optional
        Max number of rows to display. The default is None.
    max_columns : (int), optional
        Max number of columns to display. The default is None.
    max_colwidth : (int), optional
        Max column width to display. The default is None.
    """
    with pd.option_context('display.max_rows', max_rows, 'display.max_columns', max_columns,
                           'display.max_colwidth', max_colwidth):  # more options can be specified also
        print(df)
    return


def set_display_df(reset=False,max_rows=None,max_columns=None,max_colwidth=None):
    """
    Set option for DataFrame display.

    Parameters
    ----------
    reset : (bool), optional
        Reset options to default. The default is False.
    max_rows : (int), optional
        Max number of rows to display. The default is None.
    max_columns : (int), optional
        Max number of columns to display. The default is None.
    max_colwidth : (int), optional
        Max column width to display. The default is None.
    """
    if reset:
        pd.reset_option('display.max_rows')
        pd.reset_option('display.max_columns')
        pd.reset_option('display.max_colwidth')
    else:
        pd.set_option('display.max_columns', max_rows)  
        pd.set_option('display.max_rows', max_columns)  
        pd.set_option('display.max_colwidth', max_colwidth)
    return


def explore_packages(path):
    """
    Return package names contained in a path
    """
    names = []
    for sm in pkgutil.walk_packages([path]):
        names.append(sm[1])
        
    return names
    
def explore_pynter_packages():
    path = pynter.__path__[0]
    return explore_packages(path)            


def get_content_from_url(file_url):
    """
    Get content from URL.

    Parameters
    ----------
    file_url: URL to retrieve content from.

    Returns
    -------
    String with content.
    """
    import urllib.request
    with urllib.request.urlopen(file_url) as url:
       content = url.read().decode('utf-8')
    return content
    

def get_object_feature(obj,feature):
    """
    Get value of attribute or method of a generic Object.
    If feature is a single method only the string with the method's name is required.
    If the target feature is stored in a dictionary (or dict of dictionaries), a list of this format needs to be provided:
        ["method_name",key1,key2,...] - This will identify the value of Object.method[key1][key2][...] .

    Parameters
    ----------
    obj : (object)
        Generic object.
    feature : (str or list)
        Method or attribute of class for which the value is needed.
    """
    if isinstance(feature,list):
        method = feature[0]
        try:
            attr = getattr(obj,method) ()
        except:
            attr = getattr(obj,method)                
        for k in feature[1:]:
            try:
                v = attr[k]
            except:
                v = None
            if isinstance(v,dict):
                if feature.index(k) + 1 == len(feature):
                    return v
                else:
                    attr = v
            else:
                return v
            
    else:
        met = feature
        try:
            attr = getattr(obj,met) ()
        except:
            attr = getattr(obj,met)
        return attr


def get_object_from_json(object_class,path_or_string):
    """
    Build class object from json file or string. The class must posses the 'from_dict' method.

    Parameters
    ----------
    object_class : (class)
        Class of the object to decoder.
    path_or_string : (str)
        If an existing path to a file is given the object is constructed reading the json file.
        Otherwise it will be read as a string.

    Returns
    -------
    Decoded object.
    """
    if op.isfile(path_or_string):
        with open(path_or_string) as file:
            d = json.load(file)
    else:
        d = json.loads(path_or_string)

    return object_class.from_dict(d)


def save_object_as_json(object,path,sanitize=False,cls=MontyEncoder):
    """
    Save class object as json string or file. The class must posses the 'as_dict' method.

    Parameters
    ----------
    object: object of a class
    path : (str)
        Path to the destination file.  If None a string is exported.

    Returns
    -------
    d : (str)
        If path is not set a string is returned.
    """
    d = object.as_dict()
    if sanitize:
        d = jsanitize(d)
    if path:
        with open(path,'w') as file:
            json.dump(d,file,cls=cls)
        return
    else:
        return json.dumps(d,cls=cls) 


def select_objects(objects,mode='and',exclude=False,functions=None,**kwargs):
    """
    Filter objects based on different criteria. Returns a list of objects.

    Parameters
    ----------
    objects : (list)
        List of objects.
    mode : (str), optional
        Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
    exclude : (bool), optional
        Exclude the entries satisfying the criteria instead of selecting them. The default is False.
    functions : (list), optional
        Functions containing criteria. The functions must take the object as the argument and
        return a bool. 
    **kwargs : (dict)
        Keys are methods/attributes the objects have. Values contain the criteria. 
        To address more than one condition relative to the same attribute,
        use lists or tuples (e.g. charge=[0,1]).

    Returns
    -------
    output_objects : (list)
        List with selected objects.
    """    
    selected_objects = []
    entered_selection = False
    filtered_objects = objects.copy()
    if functions:
        for func in functions:
            if func:
                entered_selection = True
                if mode=='and':
                    if selected_objects:
                        filtered_objects = selected_objects.copy()
                        selected_objects = []
                for obj in filtered_objects:
                    if func(obj) == True:
                        if obj not in selected_objects:
                            selected_objects.append(obj)
    
    for key in kwargs:
        if mode=='and':
            if selected_objects or entered_selection is True:
                filtered_objects = selected_objects.copy()
                selected_objects = []
        for obj in filtered_objects:
            feature = get_object_feature(obj,key)
            if type(kwargs[key]) in [list,tuple]:
                for value in kwargs[key]:
                    if feature == value:
                        if obj not in selected_objects:
                            selected_objects.append(obj)
            elif feature == kwargs[key]:
                if obj not in selected_objects:
                    selected_objects.append(obj)  
    
    output_objects = []
    for obj in objects:
        if exclude:
            if obj not in selected_objects:
                output_objects.append(obj)
        else:
            if obj in selected_objects:
                output_objects.append(obj)    
    
    return output_objects



def sort_objects(objects,features,reverse=False):
    """
    Sort objects based on a list of features (attributes or methods of the objects, or functions)

    Parameters
    ----------
    objects : (list)
        List of objects to sort.
    features : (list)
        List of features (see get_object_feature).
    reverse : (bool)
        Reverse order.

    Returns
    -------
    (list)
        Sorted objects.
    """
    def criteria(obj):
        criteria = []
        for feature in features:
            if callable(feature):
                criteria.append(feature(obj))
            else:
                criteria.append(get_object_feature(obj,feature))
        return criteria
    sort_function = lambda obj : criteria(obj)
    sorted_objects = sorted(objects,key=sort_function,reverse=reverse)
    
    return sorted_objects


def grep(search_string,file):    
    '''
    Function that emulates "grep" function and returns a list of lines
    that contain target string
    '''         
    lines = []    
    with open (file,'r') as origin_file:
    # searching lines in input file   
        for line in origin_file:
       # emulating 'grep' command
            target_line = re.findall(search_string, line)
            if target_line:
                lines.append(line)
        return lines
         

def grep_list(search_string,target_list):
    """
    Search string in a list of strings. Returns lines that contain the searched string.
    """
    lines=[]
    for l in target_list:
        if search_string in l:
            lines.append(l)
            
    return lines



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 11 12:15:47 2021

@author: lorenzo
"""

import numpy as np
import matplotlib.pyplot as plt
from pynter.phase_diagram.chempots import PDHandler

             
class PDPlotterAdder:
    
    def __init__(self,phase_diagram=None,size=1):
        """
        Class with tools to add features to default PD plots generated by Pymatgen.

        Parameters
        ----------
        phase_diagram : (PhaseDiagram)
            Pymatgen PhaseDiagram object
        size : (float)
            Multiplier for the size of the objects added in the plot
        """
        self.pd = phase_diagram if phase_diagram else None
        self.pdh = PDHandler(phase_diagram) if phase_diagram else None
        self.size = size
        
    
    def add_points(self,points,size=1,label_size=1,color=[],edgecolor='k',label_color='k',linewidths=3,**kwargs):
        """
        Add points to plot.

        Parameters
        ----------
        points : (dict)
            Dictionary with points labels as keys and tuples or list with coordinates as values.
        size : (float)
            Float multiplier for points size. Default is 1, which would yield a default size of 450*self.size
        label_size : (float)
            Float multiplier for labels size. Default is 1, which would yield a default size of 30*self.size
        color : Color of filling of points
        edgecolor : Color of point edge
        label_color: Color of labels
        linewidths : line width of point edge
        kwargs: kwargs to pass to matplotlib

        Returns
        -------
        plt : Matplotlib object
        """
        for p in points:
            plt.scatter(points[p][0],points[p][1], color=color, edgecolor=edgecolor, linewidths=linewidths, s=450*self.size*size,**kwargs)
            plt.text(points[p][0]+(0.1/self.size*label_size),points[p][1],p,size=30*self.size*label_size,color=label_color)
        return plt
    
    
    def add_constant_chempot_line(self, comp, variable_element, chempots_ref,**kwargs):
        """
        Add line of constant chemical potential (at a given composition) to the plot. Only works for 3 component PD.

        Parameters
        ----------
        comp : (Pymatgen Composition object)
            Composition of the phase.
        variable_element : (str)
            Element chosen as indipendent variable.
        chempots_ref : (dict)
            Dictionary with fixed chemical potentials (values relative to reference phase). the format is {Element:chempot}
        **kwargs : 
            kwargs passed to Matplotlib plot function.
        Returns
        -------
        plt : Matplotlib object
        """
        axes = plt.gca()
        xlim , ylim = axes.get_xlim() , axes.get_ylim()
        plt.xlim(xlim)
        plt.ylim(ylim)
        mu = np.arange(xlim[0]-1,xlim[1]+1,0.01)
        plt.plot(mu,self.constant_chempot_line(mu,comp,variable_element,chempots_ref),
                 linewidth= 4.5*self.size , **kwargs)
        return plt
    

    def add_heatmap(self,comp,elements,cbar_label='$\Delta\mu_{O}$',cbar_values=True,**kwargs):
        """
        Add heatmap that shows the value of the last chemical potential based on the values of the other two "free" 
        chemical potentials and the composition of interest. Currently works only for 3 component PDs.

        Parameters
        ----------
        comp : (Pymatgen Composition object)
            Composition of interest to compute the chemical potential.
        elements : (list)
            List of strings with elements with free chemical potentials.
        cbar_label : (string), optional
            String with label of the colormap. The default is ''.
        cbar_values : (tuple or bool), optional
            Show max e min chempot values on colorbar. If tuple the values are used, if not the 
            minimum chempot and 0 are used. The default is True.
        **kwargs : (dict)
            kwargs for "pcolormesh" function.

        Returns
        -------
        Matplotlib object
        """        
        el1,el2 = elements  
        
        def f(mu1,mu2):            
            return self.pdh.calculate_single_chempot(comp,{el1:mu1,el2:mu2})
        
        axes = plt.gca()
        xlim , ylim = axes.get_xlim() , axes.get_ylim()
        npoints = 100
        x = np.arange(xlim[0],xlim[1]+0.1,abs(xlim[1]+0.1-xlim[0])/npoints)
        y = np.arange(ylim[0],ylim[1]+0.1,abs(ylim[1]+0.1-ylim[0])/npoints)   
        
        X,Y = np.meshgrid(x,y)
        Z = f(X,Y)

        plt.pcolormesh(X,Y,Z,vmax=0,shading='auto',**kwargs)

        cbar = plt.colorbar()
       # cbar.ax.tick_params(labelsize='xx-large')
        if cbar_values:
            if isinstance(cbar_values,tuple):
                cbar_min,cbar_max = cbar_values[0], cbar_values[1]
            else:
                cbar_min = np.around(Z.min(),decimals=1)        # colorbar min value - avoid going out of range
                cbar_max = 0                                    # colorbar max value
            plt.text(0.81,1.6,str(cbar_max),size=15)         # easier to show cbar labels as text
            plt.text(0.73,-14.2,str(cbar_min),size=15)
        cbar.set_ticks([]) # comment if you want ticks
        cbar.ax.set_yticklabels('') # comment if you want tick labels
        cbar.ax.set_ylabel(cbar_label,fontsize='xx-large')

        return plt
        

    def add_reservoirs(self,reservoirs,elements,size=1,label_size=1,color=[],edgecolor='k',
                       label_color='k',linewidths=3,**kwargs):
        """
        Parameters
        ----------
        reservoirs : (Reservoirs)
            Reservoirs object.
        elements : (list)
            List with strings of the elements to be used as free variables.
        size : (float)
            Float multiplier for points size. Default is 1, which would yield a default size of 450*self.size
        label_size : (float)
            Float multiplier for labels size. Default is 1, which would yield a default size of 30*self.size
        color : Color of filling of points
        edgecolor : Color of point edge
        label_color: Color of labels
        linewidths : line width of point edge
        kwargs: kwargs to pass to matplotlib

        Returns
        -------
        plt : Matplotlib object
        """         
        points = {}
        for r,mu in reservoirs.items():
            points[r] = [mu[el] for el in elements]
        
        return self.add_points(points,size,label_size,color,edgecolor,label_color,linewidths,**kwargs)
        
    
    def constant_chempot_line(self, mu, comp, variable_element, chempots_ref):
        """
        Function that expresses line of constant chemical potential of a given composition. Only works for 3-component PD.

        Parameters
        ----------
        mu : (float)
            Indipendent variable of chemical potential
        comp : (Pymatgen Composition object)
            Composition of the phase.
        variable_element : (Pymatgen Element object)
            Element chosen as indipendent variable.
        chempots_ref : (dict)
            Dictionary with fixed chemical potentials (values relative to reference phase). the format is {Element:chempot}
        """
        chempots_ref[variable_element] = mu
        return self.chempots_analysis.calculate_single_chempot(comp,chempots_ref)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 14 14:48:55 2021

@author: villa
"""
import warnings
from monty.json import MSONable
from monty.json import jsanitize
import numpy as np

from pynter.defects.analysis import DefectConcentrations
import copy
import os.path as op
import json
import os


class Conductivity:
    """
    Class that handles conductivity calculations.
    """
    def __init__(self,mobilities):
        """
        Parameters
        ----------
        mobilities : (dict)
            Dictionary with mobility values for the defect species. 
            Keys must contain "electrons", "holes" and the defect species names.
        """
        self.mobilities = mobilities
        
        
    def get_conductivity(self,carrier_concentrations,defect_concentrations):
        """
        Calculate conductivity from the concentrations of electrons, holes and defects and their mobilities.
        
        Parameters
        ----------
        carrier_concentrations : (list)
            List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
        defect_concentrations : (list)
            Defect concentrations in the same format as the output of DefectsAnalysis in cm^-3. 

        Returns
        -------
        sigma : (float)
            Conductivity in S/m.

        """
        e = 1.60217662e-19
        mob = self.mobilities
        cc = carrier_concentrations
        sigma_el = e * (mob['holes']*cc[0]+ mob['electrons']*cc[1])
        dc = defect_concentrations
        sigma_ionic = 0
        for d in dc:
            dname = d['name']
            if dname in mob.keys():
                sigma_ionic += mob[dname] * d['conc'] * abs(d['charge']) * e *1e06 #concentrations need to be in 1/m**3
        sigma = sigma_el + sigma_ionic
        
        return sigma



class DefectThermodynamics:
    """
    Class that handles the analysis of defect equilibrium.
    """
    
    def __init__(self,defects_analysis,bulk_dos,fixed_concentrations=None,
                 external_defects=[],xtol=1e-05):
        """
        Parameters
        ----------
        defects_analysis :
            DefectsAnalysis object.
        bulk_dos : 
            Pymatgen Dos object.
        fixed_concentrations: (dict)
            Dictionary with fixed concentrations. Keys are defect entry names in the standard
            format, values are the concentrations. (ex {'Vac_Na':1e20}) 
        external_defects : (list)
            List of external defect concentrations (not present in defect entries).
        xtol: Tolerance for bisect (scipy) to solve charge neutrality. The default is 1e-05.
        """
        self.da = defects_analysis
        self.bulk_dos = bulk_dos
        self.fixed_concentrations = fixed_concentrations if fixed_concentrations else None
        self.external_defects = external_defects if external_defects else []
        self.xtol = xtol
        

    def get_conductivities_from_thermodata(self,thermodata,mobilities):
        """
        Calculate conductivities from defect and carrier concentrations in
        thermodynamic data.

        Parameters
        ----------
        thermodata: (ThermoData)
            ThermoData object that contains the thermodynamic data.
        mobilities : (dict)
            Dictionary with mobility values for the defect species. 
            Keys must contain "electrons", "holes" and the defect specie name.
            Only the defect species which are present in this dict are 
            considered for the calculation of the conductivity.

        Returns
        -------
        conductivities: (list)
            List with conductivity values in S/m.
        """
        conductivities = []
        for i in range(0,len(thermodata.carrier_concentrations)):
            carrier_concentrations = thermodata.carrier_concentrations[i]
            dc = thermodata.defect_concentrations[i] 
            defect_concentrations = []
            for d in dc:
                if d['name'] in mobilities.keys():
                    defect_concentrations.append(d)
            conductivity = Conductivity(mobilities).get_conductivity(
                                                carrier_concentrations=carrier_concentrations,
                                                defect_concentrations=defect_concentrations)
            conductivities.append(conductivity)
        
        return conductivities 


    def get_pO2_thermodata(self,reservoirs,temperature=None,name=None):
        """
        Calculate defect and carrier concentrations as a function of the oxygen partial pressure.

        Parameters
        ----------
        reservoirs : (dict, Reservoirs or PressureReservoirs)
            Object with partial pressure values as keys and chempots dictionary as values.
        temperature : (float), optional
            Temperature in Kelvin. If None reservoirs.temperature is used. The default is None.
        name : (str), optional
            Name to assign to ThermoData.

        Returns
        -------
        Thermodata : (ThermoData)
            ThermoData object that contains the thermodynamic data:
                partial_pressures : (list)
                    List of partial pressure values.
                defect_concentrations : (list)
                    List of DefectConcentrations objects
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
                fermi_levels : (list)
                    list of Fermi level values
        """
        res = reservoirs
        if temperature:
            T = temperature
        elif hasattr(res,'temperature'):
            T = res.temperature
        else:
            raise ValueError('Temperature needs to be provided or to be present ad attribute in PressureReservoirs object')
        partial_pressures = list(res.keys())
        defect_concentrations = []
        carrier_concentrations = []
        fermi_levels=[]
        dos = self.bulk_dos
        fixed_df = self.fixed_concentrations
        ext_df = self.external_defects
        for r,mu in res.items():
            mue = self.da.solve_fermi_level(chemical_potentials=mu,bulk_dos=dos,temperature=T,
                                            fixed_concentrations=fixed_df,
                                            external_defects=ext_df,xtol=self.xtol)
            conc = self.da.defect_concentrations(mu,T,mue,fixed_df)
            carriers = self.da.carrier_concentrations(dos,temperature=T,fermi_level=mue)
            defect_concentrations.append(conc)
            carrier_concentrations.append(carriers)
            fermi_levels.append(mue)
        
        thermodata = {}
        thermodata['partial_pressures'] = partial_pressures 
        thermodata['defect_concentrations'] = defect_concentrations 
        thermodata['carrier_concentrations'] = carrier_concentrations
        thermodata['fermi_levels'] = fermi_levels
        
        thermodata = ThermoData(thermodata,temperature=temperature,name=name)
        
        return thermodata


    def get_pO2_quenched_thermodata(self,reservoirs,initial_temperature,final_temperature,
                                  quenched_species=None,quench_elements=False,name=None):
        """
        Calculate defect and carrier concentrations as a function of oxygen partial pressure with quenched defects. 
        It is possible to select which defect species to quench and which ones are free to equilibrate.
        Frozen defects in the inputs of the class are still considered.

        Parameters
        ----------
        reservoirs : (dict, Reservoirs or PressureReservoirs)
            Object with partial pressure values as keys and chempots dictionary as values.
        initial_temperature : (float)
            Value of initial temperature (K).
        final_temperature : (float)
            Value of final temperature (K).
        quenched_species : (list), optional
            List of defect species to quench. If None all defect species are quenched.The default is None.
        quench_elements : (bool)
            If True the total concentrations of elements at high temperature go in the charge neutrality at low temperature.
            If False the quenched concentrations are the ones of single defect species (e.g. elements are not allowed
            to equilibrate on different sites). The default is False.
        name : (str), optional
            Name to assign to ThermoData.
            
        Returns
        -------
        Thermodata : (thermoData)
            ThermoData object that contains the thermodynamic data:
                partial_pressures : (list)
                    List of partial pressure values.
                defect_concentrations : (list)
                    List of DefectConcentrations objects
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
                fermi_levels : (list)
                    list of Fermi level values
        """        
        T1 = initial_temperature
        T2 = final_temperature
        res = reservoirs
        if hasattr(res,'temperature'):
            if res.temperature != T1:
                warnings.warn('PressureReservoirs temperature is not set to the initial quenching temperature',UserWarning)
        partial_pressures = list(res.keys())
        fermi_levels = []
        final_defect_conc = []
        final_carrier_conc = []
            
        dos = self.bulk_dos
        fixed_df = self.fixed_concentrations
        ext_df = self.external_defects
                
        for r,mu in res.items():
            mue = self.da.solve_fermi_level(chemical_potentials=mu,bulk_dos=dos,temperature=T1,
                                            fixed_concentrations=fixed_df,
                                            external_defects=ext_df,xtol=self.xtol)
            if quench_elements:
                c1 = self.da.defect_concentrations(mu,T1,mue,fixed_df).elemental
            else:
                c1 = self.da.defect_concentrations(mu,T1,mue,fixed_df).total
            if quenched_species is None:
                quenched_concentrations = c1.copy()
            else:
                quenched_concentrations = copy.deepcopy(fixed_df) if fixed_df else {}
                for k in quenched_species:
                    quenched_concentrations[k] = c1[k]
            quenched_mue = self.da.solve_fermi_level(chemical_potentials=mu,bulk_dos=dos,temperature=T2,
                                            fixed_concentrations=quenched_concentrations,
                                            external_defects=ext_df,xtol=self.xtol)
            fermi_levels.append(quenched_mue)
            
            conc = self.da.defect_concentrations(mu,T2,quenched_mue,quenched_concentrations)
            carriers = self.da.carrier_concentrations(dos,temperature=T2,fermi_level=quenched_mue)
            final_defect_conc.append(conc)
            final_carrier_conc.append(carriers)
                
            
        thermodata = {}
        thermodata['partial_pressures'] = partial_pressures
        thermodata['fermi_levels'] = fermi_levels
        thermodata['defect_concentrations'] = final_defect_conc
        thermodata['carrier_concentrations'] = final_carrier_conc
        
        thermodata = ThermoData(thermodata,temperature=(initial_temperature,final_temperature),name=name)
        
        return thermodata


    def get_variable_species_thermodata(self,variable_defect_specie,concentration_range,chemical_potentials,
                           temperature,npoints=50,name=None):
        """
        Calculate defect and carrier concentrations as a function of the concentration of a particular 
        defect species (usually a dopant).

        Parameters
        ----------
        variable_defect_specie : (str)
            Name or element of the variable defect species.
        concentration_range : (tuple or list)
            Logaritmic range of the concentration of the variable species in cm^-3 (ex. [1,20]).
        chemical_potentials : (Chempots)
            Chempots object containing chemical potentials.
        temperature : (float)
            Temperature.
        npoints : (int), optional
            Number of points to divide concentration range. The default is 50.
        name : (str), optional
            Label for ThermoData. The default is None.

        Returns
        -------
        thermodata : (ThermoData)
            ThermoData object that contains the thermodynamic data:
                variable_defect_specie : (str)
                    Name of variable defect species.
                variable_concentrations : (list)
                    List of concentrations of variable species. 
                defect_concentrations : (list or dict)
                    Defect concentrations in the same format as the output of DefectsAnalysis. 
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
                fermi_levels : (list)
                    List of Fermi level values.
        """
        carrier_concentrations = []
        defect_concentrations = []
        fermi_levels = []
        
        concentrations = np.logspace(start=concentration_range[0],stop=concentration_range[1],num=npoints)
        for c in concentrations:
            fixed_df = self.fixed_concentrations.copy() if self.fixed_concentrations else {}
            fixed_df.update({variable_defect_specie:c})
            
            mue = self.da.solve_fermi_level(chemical_potentials=chemical_potentials,bulk_dos=self.bulk_dos,
                                            temperature=temperature,fixed_concentrations=fixed_df,
                                            external_defects=self.external_defects,xtol=self.xtol)
            
            conc = self.da.defect_concentrations(chemical_potentials=chemical_potentials,temperature=temperature,
                                                 fermi_level=mue,fixed_concentrations=fixed_df)
            
            carriers = self.da.carrier_concentrations(self.bulk_dos,temperature=temperature,fermi_level=mue)
            
            defect_concentrations.append(conc)
            carrier_concentrations.append(carriers)
            fermi_levels.append(mue)
            
        data = {}
        data['variable_defect_specie'] = variable_defect_specie
        data['variable_concentrations'] = concentrations
        data['defect_concentrations'] = defect_concentrations
        data['carrier_concentrations'] = carrier_concentrations
        data['fermi_levels'] = fermi_levels
        
        thermodata = ThermoData(data,temperature=temperature,name=name)
        
        return thermodata


    def get_variable_species_quenched_thermodata(self,
                                  variable_defect_specie,concentration_range,
                                  chemical_potentials,initial_temperature,final_temperature,
                                  quenched_species=None,quench_elements=False,npoints=50,name=None):
        """
        Calculate defect and carrier concentrations as a function of the concentration of a particular 
        defect species (usually a dopant) with quenched defects.
        It is possible to select which defect species to quench and which ones are free to equilibrate.
        Frozen defects in the inputs of the class are still considered.

        Parameters
        ----------
        variable_defect_specie : (str)
            Name or element of the variable defect species.
        concentration_range : (tuple or list)
            Logaritmic range of the concentration of the variable species in cm^-3 (ex. [1,20]).
        chemical_potentials : (Chempots)
            Chempots object containing chemical potentials.
        initial_temperature : (float)
            Value of initial temperature (K).
        final_temperature : (float)
            Value of final temperature (K).
        quenched_species : (list), optional
            List of defect species to quench. If None all defect species are quenched.The default is None.
        quench_elements : (bool)
            If True the total concentrations of elements at high temperature go in the charge neutrality at low temperature.
            If False the quenched concentrations are the ones of single defect species (e.g. elements are not allowed
            to equilibrate on different sites). The default is False.
        npoints : (int), optional
            Number of points to divide concentration range. The default is 50.
        name : (str), optional
            Name to assign to ThermoData.

        Returns
        -------
        Thermodata : (thermoData)
            ThermoData object that contains the thermodynamic data:
                variable_defect_specie : (str)
                    Name of variable defect species.
                variable_concentrations : (list)
                    List of concentrations of variable species. 
                defect_concentrations : (list or dict)
                    Defect concentrations in the same format as the output of DefectsAnalysis. 
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
                fermi_levels : (list)
                    List of Fermi level values.
        """
        mu = chemical_potentials
        T1 = initial_temperature
        T2 = final_temperature

        concentrations = np.logspace(start=concentration_range[0],stop=concentration_range[1],num=npoints)
        fermi_levels = []
        final_defect_conc = []
        final_carrier_conc = []
            
        dos = self.bulk_dos
        fixed_df = self.fixed_concentrations.copy() if self.fixed_concentrations else {}
        ext_df = self.external_defects
                
        for c in concentrations:
            fixed_df.update({variable_defect_specie:c})
            
            mue = self.da.solve_fermi_level(chemical_potentials=mu,bulk_dos=dos,temperature=T1,
                                            fixed_concentrations=fixed_df,
                                            external_defects=ext_df,xtol=self.xtol)
            if quench_elements:
                c1 = self.da.defect_concentrations(mu,T1,mue,fixed_df).elemental
            else:
                c1 = self.da.defect_concentrations(mu,T1,mue,fixed_df).total
            if quenched_species is None:
                quenched_concentrations = c1.copy()
            else:
                quenched_concentrations = copy.deepcopy(fixed_df) if fixed_df else {}
                for k in quenched_species:
                    quenched_concentrations[k] = c1[k]
            quenched_mue = self.da.solve_fermi_level(chemical_potentials=mu,bulk_dos=dos,temperature=T2,
                                            fixed_concentrations=quenched_concentrations,
                                            external_defects=ext_df,xtol=self.xtol)
            fermi_levels.append(quenched_mue)
            
            conc = self.da.defect_concentrations(mu,T2,quenched_mue,quenched_concentrations)
            carriers = self.da.carrier_concentrations(dos,temperature=T2,fermi_level=quenched_mue)
            final_defect_conc.append(conc)
            final_carrier_conc.append(carriers)
                
            
        data = {}
        data['variable_defect_specie'] = variable_defect_specie
        data['variable_concentrations'] = concentrations
        data['fermi_levels'] = fermi_levels
        data['defect_concentrations'] = final_defect_conc
        data['carrier_concentrations'] = final_carrier_conc
        
        thermodata = ThermoData(data,temperature=(initial_temperature,final_temperature),name=name)
        
        return thermodata



class ThermoData(MSONable):
    
    "Class to handle defect thermodynamics data"
    
    def __init__(self,thermodata,temperature=None,name=None):
        """
        Class that handles dict of defect thermodynamics data.
        
        Parameters
        ----------
        thermodata : (dict)
            Dict that contains the thermodynamic data, typically output from 
            ConcentrationAnalysis and PressureAnalysis class.
            Keys of the dict are set as attributes of ThermoData.
            The items of the dict usually can be:
                partial_pressures : (list)
                    List of partial pressure values.
                variable_defect_specie : (str)
                    Name of variable defect species.
                variable_concentrations : (list)
                    List of concentrations of variable species. 
                defect_concentrations : (list)
                    List of DefectConcentrations objects.
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
                conductivities : (list)
                    List of conductivity values (in S/m).
                fermi_levels : (list)
                    list of Fermi level values.
        temperature : (float), optional
            Temperature at which the data is computed. The default is None.
        name : (str), optional
            Name to assign to ThermoData. The default is None.
        """
        self.data = thermodata  
        self.temperature = temperature if temperature else None
        self.name = name if name else None
        for k,v in thermodata.items():
            setattr(self, k, v)
        
    
    def __getitem__(self,key):
        return self.data[key]
    
    def __iter__(self):
        return self.data.__iter__()
    
    def keys(self):
        return self.data.keys()
    
    def values(self):
        return self.data.values()
    
    def items(self):
        return self.data.items()
    
    
    def as_dict(self):
        """
        Returns:
            Json-serializable dict representation of ThermoData
        """
        d = {"@module": self.__class__.__module__,
              "@class": self.__class__.__name__,
              "thermodata":self.data.copy(),    
              "temperature": self.temperature,
              "name": self.name
              }
        if 'defect_concentrations' in self.data.keys():
            d['thermodata']['defect_concentrations'] = [dc.as_dict() for dc in self.defect_concentrations]
        d = jsanitize(d) #convert numpy.float64 to float
        return d        

    def to_json(self,path=None):
        """
        Save ThermoData object as json string or file

        Parameters
        ----------
        path : (str), optional
            Path to the destination file.  If None the name of ThermoData is used as filename.

        Returns
        -------
        d : (str)
            If path is not set a string is returned.
        """
        if not path:
            path = op.join(os.getcwd(),f'thermodata_{self.name}.json')
        d = self.as_dict()
        with open(path,'w') as file:
            json.dump(d,file)
        return
 

    @classmethod
    def from_dict(cls,d):
        """
        Reconstitute a ThermoData object from a dict representation created using
        as_dict().

        Args:
            d (dict): dict representation of ThermoData.

        Returns:
            ThermoData object
        """
        if 'thermodata' in d.keys():
            data = d['thermodata'].copy()
            if 'defect_concentrations' in data.keys():
                data['defect_concentrations'] = [DefectConcentrations.from_dict(dc) for dc in data['defect_concentrations']]
            temperature = d['temperature']
            name = d['name']
        else:
            data = d # recover old 
            temperature = None
            name = None
        return cls(data,temperature,name)
     
         
    @staticmethod
    def from_json(path_or_string):
        """
        Build ThermoData object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        ThermoData object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return ThermoData.from_dict(d)
            
    
    def get_specific_pressures(self,p_values):
        """
        Get ThermoData object only for specific pressure values. The closest pressure value
        present in list is chosen for each provided value.

        Parameters
        ----------
        p_values : (list)
            List of partial pressure values.

        Returns
        -------
        ThermoData object
        """
        seldata = {}
        for p in p_values:
            pressures = self.partial_pressures
            p_sel = min(pressures, key=lambda x:abs(x-p))
            index = pressures.index(p_sel)
            for k,v in self.data.items():
                for e in v:
                    if v.index(e) == index:
                        if k not in seldata.keys():
                            seldata[k] = []
                        seldata[k].append(e)
        
        temperature = self.temperature if self.temperature else None
        

        name = self.name + 'p_' + '-'.join([str(p) for p in p_values]) if self.name else None
            
        return ThermoData(seldata,temperature=temperature,name=name)
        
    
    def set_data(self,key,value):
        """
        Set data dictionary (chosen over __setitem__ to not override data by accident).
        """
        self.data[key] = value
        setattr(self, key, value)
        return




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 15:30:49 2023

@author: villa
"""

from pynter.automations.core import Automation
from pynter.automations.vasp import Schemes, NEBSchemes

def parse_common_args(parser):
    auto = Automation()
    parser.add_argument('-j','--job-script',help='Job script filename (default: %(default)s)',required=False,default=auto.job_script_filename,type=str,metavar='',dest='job_script_filename')
    parser.add_argument('-s','--status',help='Write exit status to file (default: %(default)s)',required=False,default=auto.status_filename,type=str,metavar='',dest='status_filename')
    parser.add_argument('-e','--error-check',action='store_true',help='Perform error checking (default: %(default)s)',required=False,default=False,dest='error_check')
    return parser


def setup_automation(subparsers):
    
    subparsers_automation = subparsers.add_parser('automation',help='Set up automation for calculations')
    subparsers_automation = subparsers_automation.add_subparsers()
    
    parser_vasp = subparsers_automation.add_parser('vasp',help='VASP automation')
    parser_vasp = parse_common_args(parser_vasp)
    parser_vasp.add_argument('-c','--contcar',action='store_true',help='Copy CONTCAR to POSCAR of next step',required=False,default=False,dest='contcar')
    parser_vasp.add_argument('-W','--wavecar',action='store_true',help='Copy WAVECAR to next step',required=False,default=False,dest='wavecar')
    parser_vasp.add_argument('-C','--chgcar',action='store_true',help='Copy CHGCAR to next step',required=False,default=False,dest='chgcar')
    parser_vasp.add_argument('-K','--check-kpoints',action='store_true',help='Copy WAVECAR and POSCAR only if KPOINTS of next step are the same',required=False,default=False,dest='check_kpoints')
    parser_vasp.set_defaults(func=run_automation_vasp)
    
    parser_vasp_neb = subparsers_automation.add_parser('vasp-NEB',help='VASP automation for NEB calculations')
    parser_vasp_neb = parse_common_args(parser_vasp_neb)
    parser_vasp_neb.add_argument('-c','--contcar',action='store_true',help='Copy CONTCAR to POSCAR of next step',required=False,default=False,dest='contcar')
    parser_vasp_neb.add_argument('-W','--wavecar',action='store_true',help='Copy WAVECAR to next step',required=False,default=False,dest='wavecar')
    parser_vasp_neb.add_argument('-C','--chgcar',action='store_true',help='Copy CHGCAR to next step',required=False,default=False,dest='chgcar')
    parser_vasp_neb.add_argument('-K','--check-kpoints',action='store_true',help='Copy WAVECAR and POSCAR only if KPOINTS of next step are the same',required=False,default=False,dest='check_kpoints')
    parser_vasp_neb.set_defaults(func=run_automation_vasp_neb)


def run_automation_vasp(args):
    s = Schemes(path=None,status=[], **args.__dict__)
    conv_el, conv_ionic = s.check_convergence()    
    if conv_el and conv_ionic:
        s.next_step_relaxation_schemes()
    elif s.error_check:
        s.resubmit_if_step_limits_reached()
    s.write_status()    
    
    
def run_automation_vasp_neb(args):
    s = NEBSchemes(path=None,status=[], **args.__dict__)    
    if s.is_preconvergence():
        if s.check_preconvergence_images():
            s.copy_images_next_step_and_submit()
    else:
        if s.is_NEB_job_finished():
            s.copy_images_next_step_and_submit()
    s.write_status()    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 14:52:51 2020

@author: villa
"""
import os
import os.path as op
import numpy as np
import copy

from pymatgen.core.periodic_table import Element
from pymatgen.io.vasp.inputs import VaspInput, Incar, Poscar, Kpoints

from pynter.slurm.job_settings import JobSettings
from pynter.vasp.default_inputs import DefaultInputs
from pynter.vasp.jobs import VaspJob, VaspNEBJob
from pynter.defects.defects import create_vacancies, create_substitutions


class InputSets:
    
    def __init__(self,path=None,vaspinput=None,structure=None,incar_settings=None,kpoints=None,potcar=None,
                 job_settings=None,name=None,add_parent_folder=False):
        """
        Parameters
        ----------
        path : (str)
            Main path for the scheme. If None the current work dir is used. The default is None.
        vaspinput : (Pymatgen VaspInput object), optional
            Set of VASP inputs, the default is None. if provided the other inputs of the class (structure,incar_settings,kpoints,potcar) are not needed 
        structure : (Pymatgen Structure object), optional
            Pymatgen Structure object.
        incar_settings : (Dict), optional
            Dictionary with incar flags. The default is None. If None the default settings for PBE functional from the DefaultInputs class are used.
        kpoints : (Pymatgen Kpoints object), optional
            Pymatgen Kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        potcar : (Pymatgen kpoints object), optional
            Pymatgen kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        job_settings : (Dict), optional
            Dictionary with job settings to create job script, parameters are defined in JobSettings class function. The default is None.\n
            If job_settings is None, the 'job-name' key will be added, the value is the 'name' argument if provided, if 'name' arg is \n
            None the value will be: 'no_name'.
        name : (str), optional
            Name for the system to set up scheme for. The default is None.
        add_parent_folder : (bool), optional
            Add folder to the path names like the name of the InputSets. Default is False.
        """
        self.path = op.abspath(path) if path else os.getcwd()
        if vaspinput:
            structure = vaspinput['POSCAR'].structure
            incar_settings = Incar(vaspinput['INCAR'].copy())
            kpoints = vaspinput['KPOINTS']
            potcar = vaspinput['POTCAR']

        self.structure = structure if structure else None
        if structure:
            self.incar_settings = incar_settings if incar_settings else DefaultInputs(self.structure).get_incar_default()
            self.kpoints = kpoints if kpoints else DefaultInputs(self.structure).get_kpoints_default()
            self.potcar = potcar if potcar else DefaultInputs(self.structure).get_potcar()
            if job_settings:
                self.job_settings = JobSettings(**job_settings)
            elif name:
                self.job_settings = JobSettings(slurm={'job-name':name})
            else:
                self.job_settings = JobSettings(slurm={'job-name':'no_name'})
            self.name = name if name != None else self.job_settings['job-name'] 
            
            if self.name: # this return false if name is '', the previuos line considers only if name is None
                self.job_settings['job-name'] = self.name
                
        else:
            raise ValueError('You need to provide Structure, either as Poscar object in VaspInput or in "structure" arg')

        if add_parent_folder:
            self.path = op.join(self.path,self.name)

    def __str__(self):
        printout = 'System name:"%s" \n' %self.name
        printout += 'STRUCTURE:\n'
        printout += self.structure.__str__() + '\n'
        printout += 'INCAR:\n'
        printout += '%s \n' %str(self.incar_settings)
        printout += 'KPOINTS:\n'
        printout += self.kpoints.__str__() + '\n'
        printout += 'POTCAR symbols:\n'
        printout += '%s \n' %(', '.join(self.potcar.symbols))
        printout += 'Job settings:\n'
        printout += '%s' %(str(self.job_settings))
        return printout
    
    def __repr__(self):
        return self.__str__()


    def get_vaspjob(self,setname='',pathname=''):
        """
        Generate VaspJob object from the input settings of the class.

        Parameters
        ----------
        setname : (str), optional
            String to be added to 'name' key in job_settings dictionary and to name attribute of VaspJob. The default is ''.
        pathname : (str), optional
            String to be added to self.path. The complete path will be input in 'path' arg of VaspJob. The default is ''.

        Returns
        -------
        vaspjob : (VaspJob object)
        """        
        incar_settings = self.incar_settings.copy()
        job_settings = self.job_settings.copy()
        
        incar = Incar(incar_settings)
        kpoints = self.kpoints
        poscar = Poscar(self.structure)
        potcar = self.potcar
        vaspinput = VaspInput(incar,kpoints,poscar,potcar)
        job_settings['job-name'] = '_'.join([job_settings['job-name'],setname])
        
        jobname = '_'.join([self.name,setname])
        jobpath = op.join(self.path,pathname)
        job_script_filename = job_settings['filename'] if 'filename' in job_settings.keys() else None
        vaspjob = VaspJob(path=jobpath,inputs=vaspinput,job_settings=job_settings,job_script_filename=job_script_filename,name=jobname)
        
        return vaspjob  


    def hse_bs(self,kpoints_bs=None,setname='HSE-BS',pathname='HSE-BS',**kwargs):
        """
        Set up band structure calculation with HSE
        
        Parameters
        ----------
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points.
        kwargs : (dict), optional
            Kwargs for get_kpoints_bs_default in DefaultInputs
        """
        vaspjob = self.hse_scf(setname,pathname)
        if kpoints_bs:
            kpoints = kpoints_bs
        else:
            kpoints = DefaultInputs(self.structure).get_kpoints_bs_default(hybrid_mode=True,**kwargs)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob

    def hse_dos(self,kmesh=3,setname='HSE-DOS',pathname='HSE-DOS'):
        """
        Set up DOS calculation with HSE

        Parameters
        ----------
        kmesh : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        """
        vaspjob = self.hse_scf(setname,pathname)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['ISMEAR'] = -5        
        # multiply by 3 coeff of k-mesh
        kpoints = vaspjob.inputs['KPOINTS']
        style, kpts, kpts_shift = kpoints.style, kpoints.kpts, kpoints.kpts_shift
        kpts_new = []
        kpts_new.append([k*kmesh for k in kpts[0]])
        kpoints = Kpoints(style=style,kpts=kpts_new,kpts_shift=kpts_shift)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob        

    def hse_ionic_rel(self,setname='HSE-rel',pathname='HSE-rel'):
        """
        Set up ionic relaxation with HSE
        """
        vaspjob = self.pbe_ionic_rel(setname,pathname)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3
        vaspjob.job_settings['array_size'] = 7
        return vaspjob

    def hse_ionic_vol_rel(self,setname='HSE-rel',pathname='HSE-rel'):
        """
        Relaxation of atomic positions and cell volume with HSE
        """
        vaspjob = self.hse_ionic_rel(setname,pathname)
        vaspjob.incar['ISIF'] = 3        
        return vaspjob

    def hse_ionic_rel_gamma(self,setname='HSE-rel-Gamma',pathname='HSE-rel-gamma'):
        """
        Set up HSE ionic relaxation only in gamma point 
        """
        vaspjob = self.hse_ionic_rel(setname,pathname)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))         
        return vaspjob 

    def hse_scf(self,setname='HSE-SCF',pathname='HSE-SCF'):
        """
        Set up SCF calculation with HSE
        """
        vaspjob = self.pbe_scf(setname,pathname)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3  
        vaspjob.job_settings['time'] = '72:00:00'
        return vaspjob

    def hse_scf_gamma(self,setname='HSE-SCF-Gamma',pathname='HSE-SCF-Gamma'):
        """
        Set up HSE-SCF calculation only in gamma point 
        """
        vaspjob = self.hse_scf(setname,pathname)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))        
        return vaspjob 

    def hse_standard(self,setname='HSE',pathname='HSE'):
        """
        Set for standard HSE caculation
        """
        vaspjob = self.get_vaspjob(setname,pathname)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3 
        return vaspjob


    def hubbard(self,ldauu_dict,setname,pathname):
        """
        Set for standard PBE+U calculation. The dict {Element:U_value} is used to construct the INCAR
        """
        vaspjob = self.get_vaspjob(setname,pathname)
        vaspjob.incar['LDAU'] = '.TRUE.'
        vaspjob.incar['LDAUTYPE'] = 2
        vaspjob.incar['LDAUPRINT'] = 2
        vaspjob.incar['LDAUU'] = ' '.join([str(ldauu_dict[el]) for el in ldauu_dict])
        return vaspjob


    def pbe_bs(self,kpoints_bs=None,setname='PBE-BS',pathname='PBE-BS',**kwargs):
        """
        Set up band structure calculation with PBE
        
        Parameters
        ----------
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points.
        kwargs : (dict), optional
            Kwargs for get_kpoints_bs_default in DefaultInputs
        """
        vaspjob = self.pbe_scf(setname,pathname)
        vaspjob.incar['ICHARG'] = 11
        vaspjob.incar['LORBIT'] = 11
        vaspjob.incar['ISMEAR'] = 0
        vaspjob.incar['NEDOS'] = 2000
        if kpoints_bs:
            kpoints = kpoints_bs
        else:
            kpoints = DefaultInputs(self.structure).get_kpoints_bs_default(**kwargs)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob

    def pbe_dos(self,kmesh=3,setname='PBE-DOS',pathname='PBE-DOS'):
        """
        Set up DOS calculation with PBE

        Parameters
        ----------
        kmesh : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        """
        vaspjob = self.pbe_scf(setname,pathname)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['ISMEAR'] = -5        
        # multiply by 3 coeff of k-mesh
        kpoints = vaspjob.inputs['KPOINTS']
        style, kpts, kpts_shift = kpoints.style, kpoints.kpts, kpoints.kpts_shift
        kpts_new = []
        kpts_new.append([k*kmesh for k in kpts[0]])
        kpoints = Kpoints(style=style,kpts=kpts_new,kpts_shift=kpts_shift)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob
 
    def pbe_ionic_rel(self,setname='PBE-rel',pathname='PBE-rel'):
        """
        Relaxation of atomic positions with PBE
        """
        vaspjob = self.pbe_scf(setname,pathname)
        vaspjob.incar['NSW'] = 100
        vaspjob.incar['EDIFF'] = 1e-05
        vaspjob.incar['ISIF'] = 2        
        return vaspjob 

    def pbe_ionic_rel_gamma(self,setname='PBE-rel-Gamma',pathname='PBE-rel-gamma'):
        """
        Set up PBE ionic relaxation only in gamma point 
        """
        vaspjob = self.pbe_ionic_rel(setname,pathname)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))        
        return vaspjob 

    def pbe_ionic_vol_rel(self,setname='PBE-rel-vol',pathname='PBE-rel-vol'):
        """
        Relaxation of atomic positions and cell volume with PBE
        """
        vaspjob = self.pbe_ionic_rel(setname,pathname)
        vaspjob.incar['ISIF'] = 3        
        return vaspjob 
                
    def pbe_scf(self,setname='PBE-SCF',pathname='PBE-SCF'):
        """
        Set up standard input for electronic minimization with PBE
        """        
        vaspjob = self.get_vaspjob(setname,pathname)
        vaspjob.incar['NSW'] = 0
        vaspjob.incar['EDIFF'] = 1e-06        
        return vaspjob      

    def pbe_scf_gamma(self,setname='PBE-SCF-Gamma',pathname='PBE-SCF-Gamma'):
        """
        Set up PBE-SCF calculation only in gamma point 
        """
        vaspjob = self.pbe_scf(setname,pathname)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))         
        return vaspjob 

  
class Schemes(InputSets):
    """
    Class to generate and write input files for different calculation schemes in VASP
    """
        
    def charge_states(self,charges,locpot=True):
        """
        Generate calculation schemes for calculations of different charge states.
        Only the NELECT parameter in INCAR is varied.
        A list of desired charges (float or int) is required.
        
        Returns:
            List of VaspJob objects        
        """  
        jobs = []        
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons        
        nelect = sum([ val[el.symbol]*coeff for el,coeff in self.structure.composition.items()])
        
        for q in charges:
            if isinstance(q,float) and not q.is_integer():
                q = np.around(q,decimals=1)
            stepname = f'q{q}'
            vaspjob = self.get_vaspjob(setname=stepname,pathname=stepname)
            vaspjob.incar['NELECT'] = nelect - q 
            if locpot:
                vaspjob.incar['LVTOT'] = '.TRUE.' #most likely needed for corrections in defect calculations
            jobs.append(vaspjob)
            
        return jobs


    def convergence_encut(self,encuts=[]):
        """
        Scheme for energy cutoff convergence

        Parameters
        ----------
        encuts : (list), optional
            List of energy cutoffs to use. If not provided a range from 200 to 700 eV is used.The default is [].

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []        
        if not encuts:
            encuts = range(300,800,100)

        for ec in encuts:           
            print(ec)
            stepname = f'cutoff{ec}'
            vaspjob = self.pbe_scf(setname=stepname,pathname=stepname)            
            vaspjob.incar['ENCUT'] = ec
            jobs.append(vaspjob)
            
        return jobs
        

    def convergence_kpoints(self,kpoints_meshes=[]):
        """
        Scheme for kpoints convergence.

        Parameters
        ----------
        kpoints_meshes : (list), optional
            List of kpoints meshes to use. If not provided a range from 2x2x2 to 7x7x7 is used.The default is [].

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []        
        if not kpoints_meshes:
            kpoints_meshes = []
            for k in range(2,9):
                kpoints_meshes.append((k,k,k))

        for kmesh in kpoints_meshes:
            stepname = 'k%ix%ix%i' %(kmesh[0],kmesh[1],kmesh[2])
            vaspjob = self.get_vaspjob(setname=stepname,pathname=stepname)
            vaspjob.incar.pop('KPAR', None)
            vaspjob.incar['NSW'] = 0
            vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=kmesh)
            jobs.append(vaspjob)
            
        return jobs
            

    def dielectric_properties_complete(self,scheme_name='dielectric-properties',hybrid=False):
        """
        Set up calculations for electronic and ionic contribution in the dielectric constant.
        If hybrid is set to True the LCALCEPS method is used, else the LEPSILON is used.
        """
        jobs = []
        vaspjob = self.dielectric_properties_electronic()
        jobs.append(vaspjob)
        
        if hybrid:
            vaspjob = self.dielectric_properties_ionic_lcalceps()
        else:
            vaspjob = self.dielectric_properties_ionic_lepsilon()
        jobs.append(vaspjob)
        
        return jobs


    def dielectric_properties_electronic(self,scheme_name='eps-electronic'):
        """
        Set calculation for electronic contribution to the dielectric constant (and also dielectric function).
        Uses 'LOPTICS' method in VASP.
        """
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons
        vaspjob = self.get_vaspjob(setname='eps-el',pathname=scheme_name)    
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LOPTICS'] = '.TRUE.'              
        nelect = sum([ val[el]*self.structure.composition.as_dict()[el] for el in self.structure.composition.as_dict()]) #number of electrons
        nbands = int((nelect/2 + len(self.structure.sites)/2)*3) # 3*VASP default NBANDS (NELECT/2 +NIONS/2)
        vaspjob.incar['NBANDS'] = nbands
        
        return vaspjob
        

    def dielectric_properties_ionic_lcalceps(self,scheme_name='eps-ionic-lcalceps'):
        """
        Set calculation for ionic contribution to the dielectric constant.
        Uses 'LCALCEPS' method in VASP, combined with 'IBRION=6'. Useful for Hybrid calculations where 'LEPSILON' method does not work.
        """
        vaspjob = self.get_vaspjob(setname='eps-ion-lcal',pathname=scheme_name)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LCALCEPS'] = '.TRUE.'              
        vaspjob.incar['IBRION'] = 6
        vaspjob.incar['NSW'] = 100
 
        return vaspjob


    def dielectric_properties_ionic_lepsilon(self,scheme_name='eps-ionic-lepsilon'):
        """
        Set calculation for ionic contribution to the dielectric constant.
        Uses 'LEPSILON' method in VASP, combined with 'IBRION=8'. This method does not work with HSE functionals.
        """
        vaspjob = self.get_vaspjob(setname='eps-ion-leps',pathname=scheme_name)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LEPSILON'] = '.TRUE.'              
        vaspjob.incar['IBRION'] = 8
        vaspjob.incar['NSW'] = 100
 
        return vaspjob


    def fractional_charge_linearity(self,nelect=None):
        """
        Generate calculation scheme for occupation linearity test.
        The number of electrons are scanned from NELECT to NELECT + 1 with interval of 0.2.
        
        Parameters
        --------
        nelect: (int)
            NELECT to start from. If None the total valence electrons from Potcar are used.
        Returns:
            List of VaspJob objects
        """        
        jobs = []        
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons 
        if not nelect:
            nelect = sum([ val[el]*self.structure.composition.as_dict()[el] for el in self.structure.composition.as_dict()])
        
        for q in np.arange(0,1.2,0.2):
            q = np.around(q,decimals=1)
            stepname = f'q{q}'
            vaspjob = self.get_vaspjob(setname=stepname,pathname=stepname)
            vaspjob.incar['NELECT'] = nelect + q
            jobs.append(vaspjob)
            
        return jobs
        
    
    def hse_relaxation(self,scheme_name='HSE-rel'):
        """
        Generates calculation scheme for ionic relaxation for HSE. Steps: \n
            '1-PBE-SCF': Electronic SCF with PBE \n
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE \n
            '3-HSE-SCF': Electronic SCF with HSE \n
            '4-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE
            
        Returns:
            List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-HSE-SCF','4-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4
        vaspjob = self.hse_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
    
    
    def hse_relaxation_gamma(self,scheme_name='HSE-rel-gamma'):
        """
        Generates calculation scheme for structure relaxations for HSE with more intermediate steps. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF with PBE only in Gamma point \n
            '2-PBE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE in Gamma point \n
            '3-HSE-SCF-Gamma': Electronic SCF with HSE only in Gamma point \n
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE in Gamma point \n
            '5-PBE-SCF': Electronic SCF with PBE \n
            '6-HSE-SCF': Electronic SCF with HSE \n
            '7-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE'''

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-HSE-SCF-Gamma',
                  '4-HSE-OPT-Gamma','5-PBE-SCF','6-HSE-SCF','7-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.hse_ionic_rel_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5 
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 6 
        vaspjob = self.hse_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 7
        vaspjob = self.hse_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
    
    
    def hse_relaxation_gamma_extended(self,scheme_name='HSE-rel-gamma-ext'):
        """
        Generates calculation scheme for structure relaxations for HSE with more intermediate steps. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF with PBE only in Gamma point \n
            '2-PBE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE in Gamma point \n
            '3-HSE-SCF-Gamma': Electronic SCF with HSE only in Gamma point \n
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE in Gamma point \n
            '5-PBE-SCF': Electronic SCF with PBE \n
            '6-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE \n
            '7-HSE-SCF': Electronic SCF with HSE \n
            '8-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE'''

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-HSE-SCF-Gamma',
                  '4-HSE-OPT-Gamma','5-PBE-SCF','6-PBE-OPT','7-HSE-SCF','8-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3
        vaspjob = self.hse_scf_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4
        vaspjob = self.hse_ionic_rel_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5 
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 6 
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 7 
        vaspjob = self.hse_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 8 
        vaspjob = self.hse_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
                      

    def hse_relaxation_short(self,scheme_name='HSE-rel-short'):
        """
        Generates calculation scheme for ionic relaxation for HSE (short version i.e. no PBE preliminary calculation). Steps: \n
            '1-PBE-SCF': Electronic SCF with HSE \n
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-HSE-SCF','2-HSE-OPT']        
        jobs = []
        
        sn = 1 
        vaspjob = self.hse_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.hse_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs

    
    def hse_vol_relaxation(self,scheme_name='HSE-Vrel'):
        """
        Generates calculation scheme for structure relaxations for HSE including cell volume relaxation. Steps: \n
            '1-PBE-SCF': Electronic SCF with PBE
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE
            '3-HSE-SCF': Electronic SCF with HSE
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE
            '5-HSE-VOPT': Cell volume relaxation and ionic relaxation with HSE (ISIF=3)

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-HSE-SCF','4-HSE-OPT','5-HSE-VOPT']
        jobs = []
        
        sn = 1 
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.hse_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5
        vaspjob = self.hse_ionic_vol_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        return jobs
    
    
    def hubbard_tuning(self,specie,ldauu_dict=None,u_range=(1,10),scheme_name='U_tuning'):
        """
        Generates Scheme for many calculations using different Hubbard corrections

        Parameters
        ----------
        specie : (str)
            Symbol of atomic specie to apply changes on U parameter.
        ldauu_dict : (dict), optional
            Dictionary of U parameters. Keys are pymatgen Element objects, values are the associated U parameters. The default is None.
        u_range : (Tuple), optional
            Tuple with range of U parameter to scan. The default is (1,10).
        scheme_name : (str), optional
            Name for scheme. The default is "U_tuning".

        Returns
        -------
        List of VaspJob objects
        """
        jobs = []
        
        if ldauu_dict is None:
            ldauu_dict={}
            for el in self.structure.composition.elements:
                ldauu_dict[el] = 0
        else:
            if len(ldauu_dict) != len(self.structure.composition.elements):
                raise ValueError('size of "ldauu_dict" needs to be the same as the number of species in Structure')
                
        for u in range(u_range[0],u_range[1]+1):
                        
            ldauu_dict[Element(specie)] = u
            stepname = f'U{u}'
            vaspjob = self.hubbard(ldauu_dict,setname=stepname,pathname=stepname)
            jobs.append(vaspjob)
            
        return jobs

    
    def pbe_electronic_structure(self,kmesh_dos=3, kpoints_bs=None,scheme_name='PBE-el-str'):
        """
        Generates calculation scheme for electronic structure calculations for PBE (in general non hybrid functionals) Steps: \n
            '1-PBE-relax': Relaxation of atomic positions and cell volume \n
            '2-PBE-DOS': DOS calculation with 2000 number of gridpoints on which the DOS is evaluated (NEDOS=2000) and increased k-mesh, ISMEAR is set to -5.
            '3-PBE-BS': Bandstructure calculation
        Parameters
        ----------
        kmesh_dos : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points. '
            
        Returns
        -------
        List of VaspJob objects
        """
        stepnames = ['1-PBE-relax','2-PBE-DOS','3-PBE-BS']
        jobs = []
        
        sn = 1 
        vaspjob = self.pbe_ionic_vol_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)  
        
        sn = 2 
        vaspjob = self.pbe_dos(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 3
        vaspjob = self.pbe_bs(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs

    
    def pbe_relaxation(self,scheme_name='PBE-rel'):
        """
        Generates calculation scheme for ionic relaxation with PBE functional. Steps: \n
            '1-PBE-SCF': Electronic SCF \n
            '2-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).

        Returns
        -------
        List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF','2-PBE-OPT']
        jobs = []
              
        sn = 1
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs


    def pbe_relaxation_gamma(self,scheme_name='PBE-rel-gamma'):
        """
        Generates calculation scheme for ionic relaxation with PBE functional. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF in Gamma point\n
            '2-PBE-OPT-Gamma': Ionic relaxation in Gamma point with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).
            '3-PBE-SCF': Electronic SCF \n
            '4-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).

        Returns
        -------
        List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-PBE-SCF','4-PBE-OPT']
        jobs = []

        sn = 1
        vaspjob = self.pbe_scf_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel_gamma(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
              
        sn = 3
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs
    
    
    def pbe_vol_relaxation(self,scheme_name='PBE-Vrel'):
        """
        Generates calculation scheme for ionic and cell relaxation with PBE functional. Steps: \n
            '1-PBE-SCF': Electronic SC \n
            '2-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05) \n
            '3-PBE-VOPT': Same parameters for ionic relaxation and relaxation of the cell (ISIF=3).

        Returns
        -------
        List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-PBE-VOPT']
        jobs = []
        sn = 1
        vaspjob = self.pbe_scf(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)
 
        sn = 3 
        vaspjob = self.pbe_ionic_vol_rel(setname=scheme_name+'_%i'%sn ,pathname=stepnames[sn-1])
        jobs.append(vaspjob)       
        
        return jobs
                 

class AdvancedSchemes(Schemes):
    
    
    def defects_pbe_relaxation(self,defects_with_charges,automation=False,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default defect calculation scheme with PBE.

        Parameters
        ----------
        defects_with_charges : (dict)
            List of tuples with Defect object and relative charge state list, for example [(Defect , [q1,q2,q3,...])].
        automation : (bool)
            Add default automation string: 'automation_vasp.py --contcar --chgcar --wavecar --error-check --check-kpoints'.
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []
        if automation:
            self.job_settings['add_automation'] = 'pynter automation vasp --contcar --chgcar --wavecar --error-check --check-kpoints'

        for df, charge_states in defects_with_charges:
            path = op.join(self.path,df.name)
            defect_name = df.name.replace('(','_').replace(')', '') #replace label parenthesis
            schemes_q = Schemes(path=path,structure=df.defect_structure,incar_settings=self.incar_settings,
                                job_settings=self.job_settings,name=self.name+'_'+defect_name,add_parent_folder=False)
            charge_jobs = schemes_q.charge_states(charge_states,locpot)
            for jq in charge_jobs:
                schemes_rel = Schemes(path=jq.path,vaspinput=jq.inputs,job_settings=jq.job_settings,name=jq.name)
                if rel_scheme == 'default':
                    rel_jobs = schemes_rel.pbe_relaxation()
                elif rel_scheme == 'gamma':
                    rel_jobs = schemes_rel.pbe_relaxation_gamma()
                for jrel in rel_jobs:
                        jobs.append(jrel)
        return jobs


    def interstitials_pbe_relaxation(self):
        """
        Since the interstitials structures neeed to be checked and selected, calculations should be set up manually 
        with defects_pbe_relaxation. If the function needs to be implemented, set up a criterion for the choice of 
        interstitials sites.
        """
        pass

            
    def substitutions_pbe_relaxation(self,elements_to_replace_with_charges,supercell_size=None,automation=False,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default substitutions calculation scheme with PBE.

        Parameters
        ----------
        elements_to_replace_with_charges : (dict)
            Dictionary which indicates the substitution types and the relative charge states list.
            Needs to be in this format:
                {'<new_el>-on-<old_el>':[q0,q1,q2]}, where <new_el> and <old_el> are the symbols of the element to be added and
                the element to be substituted with, respectively. [q0,q1,q2] is the list of corrisponding charge states for this substitution.
            The format of the substitution string is the same that is generated from the function "create_substitutiion_structures".
        supercell_size : (int or numpy array), optional
            Input for the make_supercell function of the Structure class. The default is 1 (the structure is not modified).
        automation : (bool)
            Add default automation string: 'automation_vasp.py --contcar --chgcar --wavecar --error-check --check-kpoints'.
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        defects_with_charges = []
        for key,charge_states in elements_to_replace_with_charges.items():
            elements_to_replace = {key.split('-on-')[1] : key.split('-on-')[0]}
            substitution = create_substitutions(self.structure,elements_to_replace=elements_to_replace,
                                                supercell_size=supercell_size)[0]
            defects_with_charges.append((substitution,charge_states))
        jobs = self.defects_pbe_relaxation(defects_with_charges,automation=automation,locpot=locpot,rel_scheme=rel_scheme)
        return jobs        
        

    def vacancies_pbe_relaxation(self,elements_with_charges,supercell_size=None,automation=False,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default vacancies calculation scheme with PBE.

        Parameters
        ----------
        elements_with_charges : (dict)
            Dictionary with element symbols as keys and relative charge states (int) lists as values ({'el':[-1,0,1]}).
        supercell_size : (int or numpy array), optional
            Input for the make_supercell function of the Structure class.
            If None the input structure is not modified. The default is None.
        automation : (bool)
            Add default automation string: 'automation_vasp.py --contcar --chgcar --wavecar --error-check --check-kpoints'.
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        defects_with_charges = []
        for el,charge_states in elements_with_charges.items():
            vacancy = create_vacancies(structure=self.structure,elements=[el],supercell_size=supercell_size)[0]
            defects_with_charges.append((vacancy,charge_states))
        jobs = self.defects_pbe_relaxation(defects_with_charges,automation=automation,locpot=locpot,rel_scheme=rel_scheme)
        return jobs

        
    
class NEBSchemes:
    
    def __init__(self,path,structures,incar_settings=None,kpoints=None,potcar=None,job_settings=None,name=None,add_parent_folder=False):
        """
        Parameters
        ----------
        path : (str)
            Main path for the scheme.        
        structures : (list of Pymatgen Structure objects), optional
            List of structures for reaction path.
        incar_settings : (Dict), optional
            Dictionary with incar flags. The default is None. If None the default settings for PBE functional from the DefaultInputs class are used.
        kpoints : (Pymatgen Kpoints object), optional
            Pymatgen Kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        potcar : (Pymatgen kpoints object), optional
            Pymatgen kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        job_settings : (Dict), optional
            Dictionary with job settings to create job script, parameters are defined in ScrpitHandler class function. The default is None.\n
            If job_settings is None, the 'name' key will be added, the value is the 'name' parameter if provided, if 'name' parameter is \n
            None the value will be: 'no_name'.
        name : (str), optional
            Name for the system to set up scheme for. The default is None.
        add_parent_folder : (bool), optional
            Add folder to the path names like the name of the InputSets. Default is False.
        """
        
        self.path = path
        self.structures = structures
        self.incar_settings = incar_settings if incar_settings else DefaultInputs(self.structures[0]).get_incar_default()
        self.kpoints = kpoints if kpoints else DefaultInputs(self.structures[0]).get_kpoints_default()
        self.potcar = potcar if potcar else DefaultInputs(self.structures[0]).get_potcar()
        if job_settings:
            self.job_settings = JobSettings(**job_settings)
        elif name:
            self.job_settings = JobSettings(slurm={'job-name':name})
        else:
            self.job_settings = JobSettings(slurm={'job-name':'no_name'})
        self.name = name if name != None else self.job_settings['job-name'] 
        
        if self.name: # this return false if name is '', the previuos line considers only if name is None
            self.job_settings['job-name'] = self.name

        if add_parent_folder:
            self.path = op.join(self.path,self.name)

        for key in ['KPAR','NCORE','NPAR']:
            if key in self.incar_settings:
                del self.incar_settings[key]

        self.images = len(self.structures)-2


    def cineb_pbe(self,nionic_steps=50,scheme_name=None,stepnames=['CINEB']):
        """
        Climbing image NEB job with PBE. The force convergence is set to 0.05 eV/A, relaxation is done with damped 
        dynamics (IBRION=3), symmetry is turned off (ISYM=0). The maximum number of ionic steps is set to nionic_steps.
        """
        scheme_name = scheme_name if scheme_name != None else 'CINEB'
        
        inputs = {}
        incar_settings = self.incar_settings.copy()
        job_settings = self.job_settings.copy()

        incar_settings.update({
            'ISYM': 0,
            'EDIFF': 1e-05,
            'EDIFFG': -0.05,
            'ISIF': 2,
            'NSW' : nionic_steps,
            'IMAGES' : self.images,
            'SPRING' : -5,
            'IOPT' : 0,
            'IBRION' : 3,
            'POTIM' : 0.05,
            'ICHAIN' : 0,
            'LCLIMB' : '.TRUE.',
            'LTANGENTOLD' : '.FALSE.',
            'LDNEB' : '.FALSE.',
            'LNEBCELL' : '.FALSE.'
            })

        incar = Incar(incar_settings)
        kpoints = self.kpoints
        potcar = self.potcar
        
        inputs['structures'] = self.structures
        inputs['INCAR'] = incar
        inputs['KPOINTS'] = kpoints
        inputs['POTCAR'] = potcar
        
        if 'add_automation' not in job_settings:
            job_settings['add_automation'] = None        
        job_settings['job-name'] = '_'.join([self.name,scheme_name])


        jobname = '_'.join([self.name,scheme_name])
        jobpath = op.join(self.path,stepnames[0])
        nebjob = VaspNEBJob(path=jobpath,inputs=inputs,job_settings=job_settings,name=jobname)

        return nebjob

 
    def neb_complete_pbe(self,nionic_steps=50,scheme_name=None):
        """
        NEB scheme with 3 steps:
            - Step 1: preconvergence (SCF) of the images with EDIFF = 1e-04
            - Step 2: NEB calculation with NSW = nionic_steps
            - Step 3: CINEB with NSW = nionic_steps
        """
        scheme_name = scheme_name if scheme_name != None else 'NEB_comp'
        jobs = []
        step1 = self.preconverge(scheme_name='NEB-SCF_1',stepnames=['1-SCF'])
        for j in step1:
            jobs.append(j)
        
        step2 = self.neb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_2',stepnames=['2-NEB'])
        jobs.append(step2)
        
        step3 = self.cineb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_3',stepnames=['3-CINEB'])
        jobs.append(step3)

        return jobs
 

    def neb_complete_4step_pbe(self,nionic_steps=50,scheme_name=None):
        """
        NEB scheme with 4 steps:
            - Step 1: preconvergence (SCF) of the images with EDIFF = 1e-04
            - Step 2: NEB calculation with NSW = 10
            - Step 3: NEB calculation with NSW = nionic_steps
            - Step 4: CINEB with NSW = nionic_steps
        """
        scheme_name = scheme_name if scheme_name != None else 'NEB_comp'
        jobs = []
        step1 = self.preconverge(scheme_name='NEB-SCF_1',stepnames=['1-SCF'])
        for j in step1:
            jobs.append(j)
        
        step2 = self.neb_pbe(scheme_name='NEB_2',stepnames=['2-NEB-NSW10'])
        step2.inputs['INCAR']['NSW'] = 10
        jobs.append(step2)

        step3 = self.neb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_3',stepnames=['3-NEB'])
        jobs.append(step3)
        
        step4 = self.cineb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_4',stepnames=['4-CINEB'])
        jobs.append(step4)

        return jobs

          
    def neb_pbe(self,nionic_steps=50,scheme_name=None,stepnames=['NEB']):
        """
        Standard NEB job with PBE. The force convergence is set to 0.1 eV/A, relaxation in done with damped 
        dynamics (IBRION=3), symmetry is turned off (ISYM=0). The maximum number of ionic steps is set to 200.
        """
        scheme_name = scheme_name if scheme_name != None else 'NEB'
        
        inputs = {}
        incar_settings = self.incar_settings.copy()
        job_settings = self.job_settings.copy()

        incar_settings.update({
            'ISYM': 0,
            'EDIFF': 1e-04,
            'EDIFFG': -0.1,
            'ISIF': 2,
            'NSW' : nionic_steps,
            'IMAGES' : self.images,
            'SPRING' : -5,
            'IOPT' : 0,
            'IBRION' : 3,
            'POTIM' : 0.05,
            'ICHAIN' : 0,
            'LCLIMB' : '.FALSE.',
            'LTANGENTOLD' : '.FALSE.',
            'LDNEB' : '.FALSE.',
            'LNEBCELL' : '.FALSE.'
            })

        incar = Incar(incar_settings)
        kpoints = self.kpoints
        potcar = self.potcar
        
        inputs['structures'] = self.structures
        inputs['INCAR'] = incar
        inputs['KPOINTS'] = kpoints
        inputs['POTCAR'] = potcar
        
        if 'add_automation' not in job_settings:
            job_settings['add_automation'] = 'automation_vasp_NEB.py'        
        job_settings['job-name'] = '_'.join([self.name,scheme_name])


        jobname = '_'.join([self.name,scheme_name])
        jobpath = op.join(self.path,stepnames[0])
        nebjob = VaspNEBJob(path=jobpath,inputs=inputs,job_settings=job_settings,name=jobname)

        return nebjob


    def preconverge(self,scheme_name=None,stepnames=['NEB-preconverge']):
        """
        Do SCF convergence for all images before starting NEB calculation
        """      
        scheme_name = scheme_name if scheme_name!=None else 'SCF'

        self.incar_settings.update({
            'EDIFF' : 1e-04,
            'NSW': 0
            })

        jobs = []
        structures = self.structures
        for s in structures:
            
            index = structures.index(s)
            image_name = str(index).zfill(2)
            
            incar = Incar(self.incar_settings)
            kpoints = self.kpoints
            potcar = self.potcar
            poscar = Poscar(s)
            vaspinput = VaspInput(incar, kpoints, poscar, potcar)
            
            job_settings = self.job_settings.copy()
            if 'add_automation' not in job_settings:
                if index ==  structures.index(structures[-1]):
                    job_settings['add_automation'] = '(cd ../ && automation_vasp_NEB.py)'
                else:
                    job_settings['add_automation'] = 'automation_vasp.py --chgcar --wavecar'
            job_settings['job-name'] = '_'.join([self.name,scheme_name,image_name])
            
            jobname = '_'.join([self.name,scheme_name,image_name])
            jobpath = op.join(self.path,stepnames[0],image_name)
            vaspjob = VaspJob(path=jobpath,inputs=vaspinput,job_settings=job_settings,name=jobname)
            jobs.append(vaspjob)
            
        return jobs      
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 29 15:16:02 2023

@author: villa
"""
import re
import os
import os.path as op
import copy

from monty.json import MSONable

from pynter import SETTINGS
from pynter.slurm.core import Slurm
from pynter.tools.utils import grep_list

class JobSettings(dict,MSONable):
    
    # switch to key value format for sbatch args
    def __init__(self,load_slurm_defaults=True,slurm=None,filename='job.sh',array_size=None,modules=None,
                 path_exe=None,add_stop_array=False,add_automation=False,add_lines_header=None,
                 add_lines_body=None,**kwargs):
        """
        Parameters
        ----------
        slurm : (Slurm) Slurm object. Handles all argments related to #SBATCH. 
            If None the defualt values are used, which will be updated with the
            user-defined **kwargs.
        array_size: (int) Number of jobs for array \n
        modules: (list) List of modules to be loaded
        path_exe: (str) Path to executable \n             
        add_stop_array : (Bool), Add lines for stopping array jobs when calculation is converged. \n                
            If True is effective only if key 'array_size' is not None
        add_automation : (str) , Automation script to add to the file.                
        add_lines_header : (List) , Lines to add in the header part of the file.
        add_lines_body : (List) , Lines to add in the body part of the file.
        **kwargs : dict. Additional kwargs which will be automatically assigned to Slurm.
        """
        super().__init__()
        
        settings = {}
        default_settings = SETTINGS['job_settings']
        
        if slurm:
            if type(slurm) is dict:
                slurm = Slurm(load_defaults=load_slurm_defaults,**slurm)
            elif slurm.__class__.__name__ != 'Slurm':
                raise ValueError('Slurm settings must be provided either as dictionary or as Slurm object')

        elif not slurm and 'slurm' not in kwargs.keys():
            slurm = Slurm(load_defaults=load_slurm_defaults,**kwargs)
        
        settings['slurm'] = slurm
    
        settings['filename'] = filename if filename is not None else default_settings['filename']
        settings['array_size'] = array_size if array_size is not None else default_settings['array_size']
        settings['modules'] = modules if modules is not None else default_settings['modules']
        settings['path_exe'] = path_exe if path_exe is not None else default_settings['path_exe']
        settings['add_stop_array'] = add_stop_array if add_stop_array else default_settings['add_stop_array']
        settings['add_automation'] = add_automation if add_automation else default_settings['add_automation']
        settings['add_lines_header'] = add_lines_header if add_lines_header is not None else default_settings['add_lines_header']
        settings['add_lines_body'] = add_lines_body if add_lines_body is not None else default_settings['add_lines_body']

        for key,value in settings.items():
            setattr(self,key,value)
            
        self.update(settings)
    
    def __getitem__(self,key):
        if key not in self.keys():
            return self.slurm[key]
        else:
            return super().__getitem__(key)
    
    def __setitem__(self,key,value):
        args = ['slurm','filename','array_size','modules','path_exe','add_stop_array',
                'add_automation','add_lines_header','add_lines_body']
        if key not in args:
            self.slurm.__setitem__(key, value)
        else:
            super().__setitem__(key,value)
        setattr(self, key, value)
        return
    
    def as_dict(self):
        d = dict(self)
        d["@module"] = type(self).__module__
        d["@class"] = type(self).__name__
        return d       
    
    def copy(self):        
        return copy.deepcopy(self)
    
    @classmethod
    def from_dict(cls,d):
        return cls(**{k: v for k, v in d.items() if k not in ["@module", "@class"]})
    
    
    @staticmethod
    def from_bash_file(path,filename='job.sh'):
        """
        Create JobSettings object from file. Cannot read added lines in header and body
        """
        file = op.join(path,filename)
        with open(file) as f:
            input_string = f.read()
        job_settings = JobSettings.from_bash_script(input_string)
        job_settings.filename = filename
        return job_settings
        
    
    @staticmethod
    def from_bash_script(input_string):
        """
        Initialize object from a bash script
        """
        lines = input_string.split('\n')
        
        slurm = Slurm.from_bash_script(input_string)
        
        string = '#SBATCH --array=1-'
        line = grep_list(string,lines)
        array_size = None
        if line:
            line = line[-1]
            line = re.sub("[^2-9]", "", line)
            array_size = int(line)
        if array_size:
            del slurm['array']
            
            
        string = 'ml '
        target_lines = grep_list(string,lines)
        if target_lines:
            modules = []
            for line in target_lines:
                if list(line[-1])[0] != '#':
                    mod_line = line.replace(string,'')
                    mod_line = mod_line.split(' ')[0] #remove space at the end
                    modules.append(mod_line) 

        string = 'srun '
        line = grep_list(string,lines)
        if line:
            line = line[-1]
            if list(line)[0] != '#':
                path_exe = line.replace(string,'')
        
        string = "if  grep -q 'Electronic convergence: True' convergence.txt  = true  && grep -q 'Ionic convergence: True' convergence.txt  = true; then"
        line = grep_list(string,lines)
        add_stop_array = False
        if line:
            if list(line)[0] != '#':
                add_stop_array = True
            
        string = 'automation'
        line = grep_list(string,lines)
        if line:
            line = line[-1]
            if list(line)[0] != '#':
                add_automation = line.lstrip() # exclude tab
            else:
                add_automation = None
        else:
            add_automation = None
            
        
        return JobSettings(load_slurm_defaults=False,slurm=slurm,array_size=array_size,
                            modules=modules,path_exe=path_exe,add_stop_array=add_stop_array,
                            add_automation=add_automation)


    def get_bash_script(self):
        lines = self.get_bash_script_header() + self.get_bash_script_body()
        string = ''.join(lines)
        return string
            
    
    def get_bash_script_body(self):
        """
        Body lines part of the job script (part after #SBATCH commands) 
        """
        f = []
        if self.array_size:
            f.append('\n')
            f.append('if [ ! -f POSCAR_initial ] ; then\n')
            f.append('    cp POSCAR POSCAR_initial\n')
            f.append('fi\n')
            f.append('if [ -f CONTCAR ]\n')
            f.append('then\n')
            f.append('    cp CONTCAR POSCAR\n') # KEEP THE TAB !)
            f.append('fi\n')
        
        f.append('\n')
        f.append('srun %s\n' %self.path_exe)

        automation_written = False
        if self.array_size:
            if self.add_stop_array:
                f.append('\n')
                f.append('pynter analysis vasprun --convergence > convergence.txt\n')
                f.append("if  grep -q 'Electronic convergence: True' convergence.txt  = true  && grep -q 'Ionic convergence: True' convergence.txt  = true; then\n")
                if self.add_automation:
                    f.append('    %s\n' %self.add_automation) # KEEP THE TAB!
                    automation_written = True
                f.append('    scancel ${SLURM_ARRAY_JOB_ID}_*\n') #KEEP THE TAB!
                f.append('fi\n')
        if self.add_automation and automation_written is False:
            f.append('\n')
            f.append('%s\n' %self.add_automation)

        if self.add_lines_body:
            for l in self.add_lines_body:
                f.append(l+'\n')

        return f
                                

    def get_bash_script_header(self):
        """
        Header lines part of the job script (part with #SBATCH commands and module loads) 
        """
        f = self.slurm.get_bash_script_lines()
        if self.array_size:
            f.append('#SBATCH --array=1-%i%%1\n' %self.array_size)
        f.append('\n')
        f.append('module purge\n')
        if self.modules:
            for m in self.modules:
                f.append(' '.join(['ml', m , '\n']))
        if self.add_lines_header:
            for l in self.add_lines_header:
                f.append(l+'\n')

        return f
    

    def write_bash_file(self,path=None,filename=None):
        """
        Write job script file.
        
        Parameters
        ----------
        path : (str), optional
            Path to write job script to. The default is None. If None work dir is used.
        filename : (str), optional
            Filename. If None self.filename is used. The default is None.
        """
        if path:
            if not os.path.exists(path):
                os.makedirs(path)
        complete_path = os.path.join(path,self.filename) if path else self.filename      
        with open(complete_path,'w') as f:
            f.write(self.get_bash_script())        
        return
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 25 14:39:37 2023

@author: villa
"""
from setuptools import setup, find_namespace_packages

with open("README.md") as file:
    long_description = file.read()

setup(
    name='pynter-defects',
    version='1.2.1',
    author='Lorenzo Villa',
    description='Tools for atomistic calculations, provides features for point-defect calculations',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(exclude=["pynter.*.tests", "pynter.*.*.tests"]),
    install_requires=[
        'ase',
        'pymatgen>=2023.3.23',
        'pymatgen-analysis-defects>=2023.3.25',
        'pymatgen-db',
        'PyYAML',
        'schedule'        
    ],
    package_data={
        "pynter":["*.yml"],
        "pynter.vasp":["*.yml"]
        },
    extra_requires={'test':'pytest'},
    entry_points={
    "console_scripts": [
        "pynter = pynter.cli.main:main"]}
)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 14:07:47 2023

@author: villa
"""

import os
import warnings
import numpy as np

from pymatgen.io.vasp.outputs import Vasprun

from pynter.tools.utils import save_object_as_json


def setup_analysis(subparsers):
    
    parser_analyse = subparsers.add_parser('analysis',help='Analysis tools for VASP')
    subparsers_analysis = parser_analyse.add_subparsers()
    
    parser_vasprun = subparsers_analysis.add_parser('vasprun',help='Analyse vasprun.xml file with pymatgen')
    setup_analyse_vasprun(parser_vasprun)
    
    
def setup_analyse_vasprun(parser_sub):    
    parser_sub.add_argument('-p','--path',help='Path to VASP calculation',required=False,type=str,default=None,metavar='',dest='path')
    parser_sub.add_argument('--bandgap',help='Analyse band gap',required=False,default=False,action='store_true',dest='bandgap') 
    parser_sub.add_argument('--convergence',help='Analyse convergence of VASP calculation',required=False,default=False,action='store_true',dest='convergence') 
    parser_sub.add_argument('--dielectric-properties',help='Analyse dielectric properties',required=False,default=False,action='store_true',dest='dielectric_properties') 
    parser_sub.add_argument('--export-dos',help='Export DOS object as json file',required=False,default=False,action='store_true',dest='export_dos') 
    parser_sub.set_defaults(func=analyse_vasprun)
    return


def analyse_vasprun(args):
    path = args.path if args.path else os.getcwd()
    vasprun = Vasprun(os.path.join(path,'vasprun.xml')) 
    
    if args.bandgap:
        analyse_bandgap(vasprun)
    if args.convergence:
        analyse_convergence(vasprun)
    if args.dielectric_properties:
        analyse_dielectric_properties(vasprun)
    if args.export_dos:
        dos = vasprun.complete_dos
        save_object_as_json(dos, 'complete_dos.json')
        
    return 
    
def analyse_bandgap(vasprun):
    gap, cbm, vbm, is_direct = vasprun.eigenvalue_band_properties
    gap = np.around(gap,decimals=4)
    print(f'Energy gap is {gap} eV')
    if is_direct:
      print('Direct gap')
    else:
      print('Indirect gap')
    print(f'CBM is at {cbm} eV')
    print(f'VBM is at {vbm} eV')
    return

def analyse_convergence(vasprun):
    conv_el = False
    conv_ionic = False
    
    try:
    	conv_el = vasprun.converged_electronic
    	conv_ionic = vasprun.converged_ionic
    except:
    	warnings.warn('Vasprun could not be read, the calculation probably failed or is not finished\n',UserWarning)
    print(f'Electronic convergence: {conv_el}')
    print(f'Ionic convergence: {conv_ionic}')
    return
    
def analyse_dielectric_properties(vasprun):
    incar = vasprun.incar
    print('')
    if 'LOPTICS' in incar:
        if incar['LOPTICS'] == True:
            print('"LOPTICS = .TRUE." found in INCAR \n')
            epsilon = vasprun.dielectric
            real_part_0 = epsilon[1][0]
            print('Real part of dielectric function at 0 frequency:')
            print(real_part_0)
            print('')
    if 'LEPSILON' in incar:
        if incar['LEPSILON'] == True:        
            print('"LEPSILON = .TRUE." found in INCAR \n')
            if incar['IBRION'] == 8:
                print('"IBRION = 8" found in INCAR \n')
            print('Macroscopic static dielectric tensor:')
            print(vasprun.epsilon_static)
            print('ionic part of the static dielectric constant. Present when it’s a DFPT run:')
            print(vasprun.epsilon_ionic)
            print('')
    return